import click
import re
from ..context import pass_context  # , on_started, on_finished
from ..actions import read_deployment_info, connect_tmux
from ..flavours import get_flavour_by_name


@click.command("connect")
@click.option("-l", "--user", default="root")
@click.option(
    "-g",
    "--geometry",
    help='Tmux geometry, 2 splitting indications are supported: +/*, examples: "1+3+2" (3 sucessive panes respectively horizontally splited by 1,3 and 2), "2*3" (2 sucessive panes horizontally splitted by 3)',
)
@click.option(
    "-d",
    "--deployment-file",
    help="Deployment file, take the latest created in deploy directory by default",
)
@click.option(
    "-f",
    "--flavour",
    help="flavour, by default it's extracted from deployment file name",
)
@click.option("-pc", "--pane-console", is_flag=True, help="Add a pane console")
@click.argument("host", nargs=-1)
@pass_context
# TODO @on_finished(lambda ctx: ctx.state.dump())
# TODO @on_started(lambda ctx: ctx.assert_valid_env())
def cli(ctx, user, host, geometry, pane_console, deployment_file, flavour):
    """Connect to host."""
    read_deployment_info(ctx, deployment_file)

    # determine flavour name
    if not flavour:
        match = re.match(r"^.*::(.+)\..*$", ctx.deployment_filename)
        if match:
            flavour = match.group(1)
        else:
            raise click.ClickException(
                "Cannot determined flavour, one must by provided (option fix deployment file name or use --flavour option"
            )

    ctx.flavour = get_flavour_by_name(flavour)(ctx)

    if not host or len(host) > 1:
        # TODO  add wait_ssh
        connect_tmux(ctx, user, host, pane_console, geometry, "nxc")
    else:
        ctx.flavour.ext_connect(user, host[0])

        # if "docker-compose-file" in ctx.deployment_info:
        #    connect_docker(ctx, user, host[0])
        # else:
        #    connect(ctx, user, host[0])



import os
import os.path as op
import json
import re
import subprocess
import sys

from io import open

import click

from ..actions import get_nix_command
from ..context import pass_context

from ..platform import platform_detection
from ..utils import copy_tree

NXC_NIX_PATH = op.abspath(op.join(op.dirname(__file__), "../../nix"))


@click.command("init")
# @click.option("-f", "--force", is_flag=True, help="Overwrite existing env")
@click.option(
    "--no-symlink",
    is_flag=True,
    help="Disable symlink creation to nxc.json (need to change directory for next command",
)
@click.option(
    "-n", "--disable-detection", is_flag=True, help="Disable platform detection."
)
@click.option(
    "-f",
    "--default-flavour",
    type=click.STRING,
    help="Set default flavour to build, if not given nixos-compose try to find a good",
)
@click.option(
    "--list-flavours-json",
    is_flag=True,
    help="List description of flavours, in json format",
)
@click.option(
    "-F", "--list-flavours", is_flag=True, help="List available flavour",
)
@click.option(
    "-t", "--template", default="basic", help="Use a template", show_default=True,
)
@click.option(
    "--use-local-templates",
    is_flag=True,
    default=False,
    help="Either use the local templates or not",
)
@click.option(
    "--list-templates-json",
    is_flag=True,
    default=False,
    help="Display the list of available templates as JSON",
)
@pass_context
# @on_finished(lambda ctx: ctx.state.dump())
def cli(
    ctx,
    no_symlink,
    disable_detection,
    default_flavour,
    list_flavours,
    list_flavours_json,
    template,
    use_local_templates,
    list_templates_json,
):
    """Initialize a new environment."""

    nix_cmd_base = get_nix_command(ctx)

    create = click.style("   create", fg="green")

    nxc_json_file = op.abspath(op.join(ctx.envdir, "nxc.json"))
    description_flavours_file = op.abspath(op.join(NXC_NIX_PATH, "flavours.json"))
    description_flavours = json.load(open(description_flavours_file, "r"))

    if list_flavours:
        for k in description_flavours.keys():
            click.echo(f"{k: <18}: {description_flavours[k]['description']}")
        sys.exit(0)

    if list_flavours_json:
        print(json.dumps(description_flavours, indent=4))
        sys.exit(0)

    if list_templates_json:
        out_file = "/tmp/.template_list.json"
        flake_location = (
            "."
            if use_local_templates
            else "git+https://gitlab.inria.fr/nixos-compose/nixos-compose"
        )
        subprocess.call(
            nix_cmd_base + ["build", f"{flake_location}#showTemplates", "-o", out_file],
            stdout=subprocess.DEVNULL,
            # stderr=subprocess.DEVNULL,
        )

        list_templates = json.load(open(out_file, "r"))
        print(json.dumps(list_templates, indent=4))
        sys.exit(0)

    if not disable_detection:
        platform_detection(ctx)

    flake_location = (
        "."
        if use_local_templates
        else "git+https://gitlab.inria.fr/nixos-compose/nixos-compose"
    )
    res = subprocess.run(
        nix_cmd_base + ["flake", "new", "-t", f"{flake_location}#{template}", "nxc"],
        capture_output=True,
    )
    # Workaround for nix 2.10.3 due to bug with flae new -t and store located in  ~/.local/share/nix/root
    # Should be resolved with availability of source tree abstraction
    # see https://github.com/NixOS/nix/pull/6530
    if res.returncode:
        # local_store_path can be present into one or three lines
        r = re.compile(".*local/share/nix/root/nix/store.*")
        res_error = res.stderr.decode()
        local_store_path = list(filter(r.match, res_error.split()))
        if local_store_path:
            copy_tree(local_store_path[0][1:-1], ctx.envdir)
            subprocess.run(["chmod", "-R", "gu+w", ctx.envdir])
        else:
            ctx.elog("Flake new from template failed:")
            ctx.elog(f"returncode: {res.returncode}")
            ctx.elog(f"stdout: {res.stdout.decode()}")
            ctx.elog(f"stderr: {res.stderr.decode()}")
            sys.exit(1)
    nxc_json = {
        "composition": "composition.nix",  # TODO to enhance
        "default_flavour": default_flavour,
    }

    if ctx.platform:
        nxc_json["platform"] = ctx.platform.name
        if not default_flavour:
            nxc_json["default_flavour"] = ctx.platform.default_flavour
        nxc_json_str = json.dumps(nxc_json)

        ctx.nxc = nxc_json
        ctx.nxc_file = nxc_json_file

        click.echo("   " + create + "  " + nxc_json_file)

        with open(nxc_json_file, "w") as f:
            f.write(nxc_json_str)

    if not no_symlink:
        os.symlink(nxc_json_file, op.abspath(op.join(ctx.envdir, "..", "nxc.json")))

    ctx.log(
        "\nInitialized nixos-compose environment in "
        + click.style(click.format_filename(ctx.envdir), fg="green")
        + "\n"
    )



import click

import time
import os

import os.path as op
import subprocess
import sys
import glob
import pyinotify
import asyncio
import re
import tempfile
import ptpython.repl
from halo import Halo

from ..context import pass_context, on_finished, on_started
from ..flavours import get_flavour_by_name

from ..actions import (
    read_test_script,
    read_hosts,
    translate_hosts2ip,
    push_on_machines,
    realpath_from_store,
    get_fs_type,
)

from ..driver.driver import Driver
from ..httpd import HTTPDaemon

machines_file_towait = ""
notifier = None


class EventHandler(pyinotify.ProcessEvent):
    def process_IN_CREATE(self, event):
        if event.pathname == machines_file_towait:
            notifier.loop.stop()


# TODO define scope of dry_run, must you dry_run deployment file creation ?
# Also how to address driver part
# def dry_run_print(ctx, launch_cmd):
#     ctx.log("Dry-run:")
#     ctx.log("   launch command:              {launch_cmd}")
@click.command("start")
@click.option(
    "-I",
    "--interactive",
    is_flag=True,
    help="drop into a python repl with driver functions",
)
@click.option(
    "-F",
    "--forward-ssh-port",
    is_flag=True,
    help="forward ssh port with nixos-test-driver forward-ssh-port",
)
@click.option(
    "-m",
    "--machines-file",
    type=click.Path(resolve_path=True),
    help="file that contains remote machines names to (duplicates are considered as one).",
)
@click.option(
    "-W", "--wait-machine-file", is_flag=True, help="wait machnes-file creation"
)
@click.option(
    "-s",
    "--ssh",
    type=click.STRING,
    default="ssh -l root ",
    help="specify particular ssh command",
)
@click.option(
    "-S",
    "--sudo",
    type=click.STRING,
    default="sudo",
    help="specify particular sudo command",
)
@click.option(
    "-p",
    "--push-path",
    help="remote path where to push image, kernel and kexec_script on machines (use to re-kexec)",
)
@click.option(
    "--reuse",
    is_flag=True,
    help="supposed a previous succeded start (w/ root access via ssh)",
)
@click.option(
    "--remote-deployment-info",
    is_flag=True,
    help="deployement info is served by http (in place of kernel parameters)",
)
@click.option(
    "-c",
    "-C",
    "--composition",
    type=click.STRING,
    help="specify composition, can specify flavour e.g. composition::flavour",
)
@click.option(
    "-f", "--flavour", type=click.STRING, help="specify flavour",
)
@click.option(
    "-t", "--test-script", is_flag=True, help="execute testscript",
)
@click.option(
    "--file-test-script", type=click.STRING, help="alternative testscript",
)
@click.option(
    "-w",
    "--sigwait",
    is_flag=True,
    help="wait any signal to exit after a start only action (not testscript execution or interactive use)",
)
@click.argument(
    "roles_quantities_file", required=False, default=None, type=click.Path(exists=True)
)
# @click.option(
#     "--dry-run", is_flag=True, help="Show what this command would do without doing it"
# )
@pass_context
@on_finished(lambda ctx: ctx.show_elapsed_time())
@on_started(lambda ctx: ctx.assert_valid_env())
def cli(
    ctx,
    interactive,
    machines_file,
    wait_machine_file,
    ssh,
    sudo,
    push_path,
    forward_ssh_port,
    reuse,
    composition,
    flavour,
    remote_deployment_info,
    test_script,
    file_test_script,
    sigwait,
    roles_quantities_file
    # dry_run,
):
    """Start Nixos Composition."""
    flavour_name = flavour
    if test_script:
        execute_test_script = True
        test_script = None
    else:
        execute_test_script = False

    if file_test_script:
        raise click.ClickException(
            "alternative test-scipt execution not yet implemented"
        )

    ctx.log("Starting")

    ctx.ssh = ssh
    ctx.sudo = sudo
    ctx.push_path = push_path
    ctx.interactive = interactive
    ctx.execute_test_script = execute_test_script
    ctx.sigwait = sigwait

    if remote_deployment_info:
        ctx.use_httpd = True

    build_path = op.join(ctx.envdir, "build")

    if not op.exists(build_path):
        raise click.ClickException(
            "You need build composition first, with nxc build command"
        )

    # Handle cases where machines list must be provided
    machines = []
    if machines_file and not op.isfile(machines_file) and not wait_machine_file:
        raise click.ClickException(f"{machines_file} file does not exist")

    if push_path and not machines_file:
        raise click.ClickException("machines_file must be provide to use push_path")

    if wait_machine_file:
        if not machines_file:
            raise click.ClickException(
                "You need to provide --machines-file option with --wait"
            )

        if not op.isfile(machines_file):
            # ctx.log(f"Waiting {machines_file} file creation")
            # TODO: add quiet option
            spinner = Halo(text=f"Waiting for {machines_file} creation", spinner="dots")
            spinner.start()

            if "nfs" == get_fs_type(machines_file):
                while not op.isfile(machines_file):
                    time.sleep(0.1)
            else:
                wm = pyinotify.WatchManager()  # Watch Manager
                loop = asyncio.get_event_loop()

                global notifier
                notifier = pyinotify.AsyncioNotifier(
                    wm, loop, default_proc_fun=EventHandler()
                )

                global machines_file_towait
                machines_file_towait = machines_file

                # TODO race condition remains possible ....
                wm.add_watch(op.dirname(machines_file), pyinotify.CREATE)
                loop.run_forever()
                notifier.stop()

            # ctx.log(f"{machines_file} file created")
            spinner.succeed(f"{machines_file} file created")

    # Determine composition and flavour name
    # First case composition is given not flavour name
    if composition and (flavour_name is None):
        splitted_composition = composition.split("::")
        len_splitted_composition = len(splitted_composition)
        if len_splitted_composition == 2:

            composition_name, flavour_name = splitted_composition
            composition_all_in_one_file = op.join(ctx.envdir, f"build/::{flavour_name}")

            if not op.lexists(composition_all_in_one_file):
                build_path = op.join(ctx.envdir, f"build/{composition}")
                if not op.lexists(build_path):
                    raise click.ClickException(
                        f"Build file does not exist: {build_path}"
                    )
            else:
                build_path = composition_all_in_one_file

            ctx.flavour = get_flavour_by_name(flavour_name)(ctx)
            ctx.composition_name = composition_name
            ctx.composition_flavour_prefix = composition
            ctx.composition_basename_file = composition_name
        else:
            raise Exception(
                "Sorry, provide only flavour or only composition is not supported"
            )

    if composition is None:
        if flavour_name:
            search_path = f"{build_path}/*::{flavour_name}"
        else:
            search_path = f"{build_path}/*"

        build_paths = glob.glob(search_path)
        if not build_paths:
            raise click.ClickException("Failed to find last build")

        last_build_path = max(
            build_paths, key=lambda x: os.stat(x, follow_symlinks=False).st_ctime,
        )

        ctx.log("Use last build:")
        ctx.glog(last_build_path)

        build_path = last_build_path
        ctx.composition_flavour_prefix = op.basename(last_build_path)

        splitted_basename = ctx.composition_flavour_prefix.split("::")

        if splitted_basename[0] == "":
            raise click.ClickException("Sorry, composition name must be provided")

        ctx.composition_name = splitted_basename[0]
        ctx.composition_basename_file = ctx.composition_name

        if flavour_name:
            assert flavour_name == splitted_basename[1]

        ctx.flavour = get_flavour_by_name(splitted_basename[1])(ctx)

    if op.isdir(build_path) and len(os.listdir(build_path)) == 0:
        ctx.wlog(f"{build_path} is an empty directory, surely a nixos-test result !")
        sys.exit(2)

    ctx.compose_info_file = realpath_from_store(ctx, build_path)
    # TODO remove only available in nixpkgs version 20.03 and before
    # if build is nixos_test result open log.html
    nixos_test_log = op.join(build_path, "log.html")
    if op.exists(nixos_test_log) and op.isfile(nixos_test_log):
        subprocess.call(f"xdg-open {nixos_test_log}", shell=True)
        sys.exit(0)

    nixos_test_driver = op.join(build_path, "bin/nixos-test-driver")
    if op.exists(nixos_test_driver) and op.isfile(nixos_test_driver):
        test_script = None
        # Deduce which nixos_test_driver is used (WARNNING: very fragile)
        with open(nixos_test_driver) as f:
            driver_script = f.read()

        after_nixos_21_05 = False

        if "startScript" in driver_script:
            after_nixos_21_05 = True
            ctx.vlog("Detected Nixos Test Driver post 21.05")
        else:
            ctx.vlog("Detected Nixos Test Driver pre 21.11")

        if machines_file:
            raise click.ClickException(
                "Nixos Driver detected, --machines-files can not by use here."
            )
        ctx.log("Nixos Driver detected")

        if not interactive:
            test_script = read_test_script(ctx, op.join(build_path, "test-script"))

        if forward_ssh_port:
            ctx.forward_ssh_port = True
            test_script = "start_all(); [m.forward_port(22022+i, 22) for i, m in enumerate(machines)]; join_all();"
            nodes = [n.split("-")[1] for n in re.findall(r"run-\w+-vm", driver_script)]

            if not ctx.compose_info:
                ctx.compose_info = {}
            ctx.compose_info["nodes"] = nodes

            ctx.flavour.generate_deployment_info()

        if "QEMU_OPTS" in os.environ:
            qemu_opts = os.environ["QEMU_OPTS"]
        else:
            qemu_opts = ""
        os.environ["QEMU_OPTS"] = f"{qemu_opts} -nographic"
        print(f"qemu_opts: {qemu_opts}")
        print(f"cmd: {nixos_test_driver}")

        if not test_script:
            ctx.wlog("Not test_script provided")

        cmd = nixos_test_driver

        if not after_nixos_21_05:
            if test_script:
                os.environ["tests"] = test_script
            subprocess.call(cmd)
        else:
            if interactive:
                cmd = f"{cmd} -I"
            if test_script:
                with tempfile.NamedTemporaryFile() as tmp:
                    ctx.vlog(f"Create temporay test_script {tmp.name}")
                    tmp.write(test_script.encode())
                    tmp.seek(0)
                    subprocess.call(f"{cmd} {tmp.name}", shell=True)
            else:
                subprocess.call(cmd, shell=True)
        sys.exit(0)

    # if not machines_file and ctx.platform:
    # TODO
    # machines = ctx.platform.retrieve_machines(ctx)
    # import pdb; pdb.set_trace()
    if ctx.platform:
        if reuse:
            (ssh, sudo, push_path) = ctx.platform.subsequent_start_values
        else:
            (ssh, sudo, push_path) = ctx.platform.first_start_values
        if ctx.push_path is None:
            ctx.push_path = push_path

    if machines_file:
        machines = read_hosts(machines_file)

    if machines:
        translate_hosts2ip(ctx, machines)
        print(ctx.ip_addresses, ctx.host2ip_address)

    if roles_quantities_file:
        ctx.load_role_quantities_file(roles_quantities_file)

    ctx.flavour.generate_deployment_info()

    if ctx.ip_addresses and (ctx.flavour.name != "vm-ramdisk"):
        if ctx.use_httpd:
            ctx.vlog("Launch: httpd to distribute deployment.json")
            ctx.httpd = HTTPDaemon(ctx=ctx)

        if hasattr(ctx.flavour, "generate_kexec_scripts"):
            ctx.flavour.generate_kexec_scripts()

        if ctx.push_path:
            push_on_machines(ctx)

        if ctx.use_httpd:
            ctx.httpd.start(directory=ctx.envdir)

        if not interactive:
            ctx.flavour.launch()
            sys.exit(0)
            # if ctx.flavour["image"]["type"] == "ramdisk":
            #     ctx.log("Launch ssh(s) kexec")
            #     launch_ssh_kexec(ctx)
            #     time.sleep(10)
            #     wait_ssh_ports(ctx)
            #     sys.exit(0)
            # if ctx.flavour_name == "g5k-image":
            #     generate_kadeploy_envfile(ctx)
            #     launch_kadeploy(ctx)
            #     sys.exit(0)

    # elif ctx.flavour_name == "docker":
    #     ctx.mode = DRIVER_MODES["docker"]
    # else:
    #     ctx.mode = DRIVER_MODES["vm"]

    # use_remote_deployment = False
    # if use_remote_deployment:
    #     httpd = HTTPDaemon()
    #     ctx.log(f"Launch httpd: port: {httpd.port}")
    #     httpd.start()

    test_script = read_test_script(ctx, ctx.compose_info)

    if forward_ssh_port:
        # ctx.mode = DRIVER_MODES["vm-ssh"]
        ctx.forward_ssh_port = forward_ssh_port
        test_script = None

    # driver(ctx, driver_repl, test_script)
    # launch_vm(ctx, deployment, 0)
    # wait_ssh_ports(ctx, ips, False)
    # httpd.stop()

    if not interactive and not execute_test_script:
        test_script = "start_all()"

    with Driver(
        # args.start_scripts, args.vlans, args.testscript.read_text(), args.keep_vm_state
        ctx,
        [],
        [],
        test_script,
        False,
    ) as driver:
        if interactive:
            ptpython.repl.embed(driver.test_symbols(), {})
        elif execute_test_script:
            tic = time.time()
            driver.run_tests()
            toc = time.time()
            ctx.glog(f"test script finished in {(toc-tic):.2f}s")
        else:
            ctx.glog("just start ???")
            driver.test_script()
    if ctx.use_httpd:
        ctx.httpd.stop()

    ctx.glog("That's All Folk")



import execo
import execo_g5k
import execo_engine
from execo import Process, Host, Remote, SshProcess, Report
from execo_g5k import get_oar_job_nodes, oarsub, oardel, OarSubmission, wait_oar_job_start
from execo_engine import Engine
import os
import os.path as op
import time
import logging
import tempfile
from .context import Context
from .actions import realpath_from_store, translate_hosts2ip
from .flavours.grid5000 import G5kRamdiskFlavour, G5KImageFlavour
from .g5k import key_sleep_script
from .httpd import HTTPDaemon

# Stolen from Adrien Faure: github.com:adfaure/vinix.git
def get_size(start_path: str):
    """Get the size of the folder `start_path`"""
    total_size = 0
    for dirpath, _, filenames in os.walk(start_path):
        for f in filenames:
            file_path = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(file_path):
                total_size += os.path.getsize(file_path)

    return total_size

# Stolen from Adrien Faure: github.com:adfaure/vinix.git
def print_total_size(store_path: str, build_node, location_nix_store, nix_chroot_script="/tmp/nix-user-chroot-companion/nix-user-chroot.sh") -> int:
    """
    Get the store reference of the store provided store path
    """
    dump_path = "/tmp/nix_store.dump"
    nix_store_cmd = f"/bin/bash {nix_chroot_script} store {store_path} {dump_path}"
    nix_store_remote = SshProcess(nix_store_cmd, build_node, shell=True)
    nix_store_remote.run()

    read_dump_remote = SshProcess(f"cat {dump_path}", build_node)
    read_dump_remote.run()
    output = read_dump_remote.stdout
    # print(output)

    total_size = 0
    # for path in output.decode().splitlines():
    for path in output.splitlines():
        real_path = path.replace("/nix/store", location_nix_store)
        total_size += get_size(real_path)

    return total_size

def clean_nix_store(build_node, nix_chroot_script):
    # 1) get and remove all the `execo builds`
    path_gcroots = f"{os.environ['HOME']}/.nix/var/nix/gcroots/auto"
    print(f"gcroots: {path_gcroots}")
    # for link in os.listdir(path_gcroots):
    #     full_link = op.join(path_gcroots, link)
    #     actual_result = os.readlink(full_link)
    #     print(f"Found: {actual_result}")
    #     # if op.basename(actual_result) == "execo_build" and op.exists(actual_result):
    #     if op.exists(actual_result):
    #         print(f"Removing: {actual_result}")
    #         os.remove(actual_result)


    # 2) running nix garbage collect
    clean_store_remote = SshProcess(f"/bin/bash {nix_chroot_script} clean", build_node)
    clean_store_remote.run()
    print(clean_store_remote.stdout)

def get_build_node(site, cluster, extra_job_type, walltime=3600):
    oar_job = oarsub([(OarSubmission(f"{{cluster='{cluster}'}}/nodes=1", walltime, job_type=["allow_classic_ssh"] + extra_job_type), site)])
    job_id, site = oar_job[0]
    wait_oar_job_start(job_id, site)
    build_node = get_oar_job_nodes(job_id, site)[0] # there is only one node
    return (job_id, build_node)

def build_derivation(build_node, nxc_path, flavour, composition_name, nix_chroot_script, clean_store):
    if nix_chroot_script is None:
        # Step 1: git clone the nix-user-chroot-companion ----------------------------------------------
        git_nix_chroot_command = "cd /tmp; /usr/bin/git clone git@github.com:GuilloteauQ/nix-user-chroot-companion.git"
        git_nix_chroot_remote = SshProcess(git_nix_chroot_command, build_node, shell=True)
        git_nix_chroot_remote.run()
        nix_chroot_script = "/tmp/nix-user-chroot-companion/nix-user-chroot.sh"

    nix_chroot_script = realpath_from_store(Context(), nix_chroot_script)

    # Step 2: execute nxc build --------------------------------------------------------------------
    result_path = f"{op.realpath(nxc_path)}/execo_build"

    if clean_store:
        if op.exists(result_path):
            clean_store_remote = SshProcess(f"/bin/bash {nix_chroot_script} delete {result_path}", build_node)
            clean_store_remote.run()
        
        execo_engine.log.logger.info("Starting Garbage Collecting")
        clean_nix_store(build_node, nix_chroot_script)
        execo_engine.log.logger.info("Done Garbage Collecting")

    execo_engine.log.logger.info("Starting Building")
    nxc_build_command = f"/bin/bash {nix_chroot_script} {nxc_path} {composition_name} {flavour} {result_path}"
    nxc_build_remote = SshProcess(nxc_build_command, build_node, shell=True)
    start_build_time = time.time()
    nxc_build_remote.run()
    end_build_time = time.time()

    execo_engine.log.logger.info(f"Done building. Composition built in {end_build_time - start_build_time} seconds")

    user = os.environ["USER"]
    location_nix_store = f"/home/{user}/.nix/store"
    sym_link_path = str(op.realpath(op.join(nxc_path, result_path)))
    compose_info_path = sym_link_path.replace("/nix/store", location_nix_store)

    execo_engine.log.logger.info("Now checking the size of the output")
    # total_output_size = print_total_size(compose_info_path, build_node, location_nix_store)
    total_output_size = print_total_size(sym_link_path, build_node, location_nix_store, nix_chroot_script=nix_chroot_script)
    execo_engine.log.logger.info(f"The total size of the output is {total_output_size} bytes")
    return (compose_info_path, end_build_time - start_build_time, total_output_size)


def build_nxc_execo(nxc_path,
                    site,
                    cluster,
                    walltime=3600,
                    flavour="g5k-ramdisk",
                    composition_name="composition",
                    extra_job_type=[],
                    nix_chroot_script=None,
                    clean_store=False):
    """
    Reserves the g5k nodes and build the composition
    returns the path to the compose_info_file
    """
    (job_id, build_node) = get_build_node(site, cluster, extra_job_type, walltime)
    execo_engine.log.logger.info(f"Building on node {build_node.address}")

    infos_build = build_derivation(build_node, nxc_path, flavour, composition_name, nix_chroot_script, clean_store)

    execo_engine.log.logger.info("Now giving back the build node")
    oardel([(job_id, site)])

    #execo_engine.log.logger.info(f"The compose info file is stored at {compose_info_path}")
    return infos_build

def get_envdir(ctx):
    if os.path.isfile("nxc.json"):
        if os.path.islink("nxc.json"):
            ctx.nxc_file = os.readlink("nxc.json")
        else:
            ctx.nxc_file = op.abspath("nxc.json")
        with open(ctx.nxc_file, "r") as f:
            ctx.load_nxc(f)

        ctx.envdir = op.dirname(ctx.nxc_file)
    else:
        raise Exception("Cannot find `nxc.json`")

def get_oar_job_nodes_nxc(oar_job_id,
                          site,
                          compose_info_file=None,
                          flavour_name="g5k-ramdisk",
                          composition_name="composition",
                          roles_quantities={}):
    """
    Brother of the "get_oar_job_nodes" function from execo
    but does the mapping with roles from NXC
    """
    ctx = Context()
    # TODO: kaberk
    ctx.composition_name = composition_name
    ctx.flavour_name = flavour_name
    ctx.set_roles_quantities(roles_quantities)

    ctx.envdir = None
    get_envdir(ctx)

    if compose_info_file:
        ctx.compose_info_file = compose_info_file
    else:
        build_folder = op.join(ctx.envdir, "build")
        simlink_build = op.join(build_folder, f"{ctx.composition_name}::{ctx.flavour_name}")
        ctx.compose_info_file = realpath_from_store(ctx, simlink_build)

    print(f"compose info file: {ctx.compose_info_file}")

    g5k_nodes = get_oar_job_nodes(oar_job_id, site)
    print(f"G5K nodes: {g5k_nodes}")
    machines = [node.address for node in g5k_nodes]
    if len(machines) > 4:
        ctx.use_http = True
        ctx.httpd = HTTPDaemon(ctx=ctx)
        ctx.httpd.start(directory=ctx.envdir)
    translate_hosts2ip(ctx, machines)

    if flavour_name == "g5k-ramdisk":
        flavour = G5kRamdiskFlavour(ctx)
    elif flavour_name == "g5k-image":
        flavour = G5KImageFlavour(ctx)
    else:
        raise Exception(f"'{flavour_name}' is not an available flavour")
    # ?!
    flavour.ctx.flavour = flavour

    flavour.generate_deployment_info()

    flavour.ctx.mode = {"name": "ssh", "vm": False, "shell": "ssh"}
    # flavour.ctx.ssh = f"OAR_JOB_ID={oar_job_id} oarsh"
    flavour.ctx.ssh = "ssh"
    flavour.ctx.sudo = "sudo-g5k"

    flavour.ctx.log("Launch ssh(s) kexec")
    if hasattr(flavour, "generate_kexec_scripts"):
        flavour.generate_kexec_scripts()
        flavour.launch()
    else:
        tmp = tempfile.NamedTemporaryFile(delete=False)
        try:
            machines_str = ""
            for machine in machines:
                machines_str += f"{machine}\n"
            tmp.write(machines_str.encode('utf-8'))
            tmp.flush()
            flavour.launch(machine_file=tmp.name)
        finally:
            tmp.close()
            os.unlink(tmp.name)

    roles = {}
    for ip_addr, node_info in flavour.ctx.deployment_info["deployment"].items():
        node_role = node_info["role"]
        if node_role in roles:
            roles[node_role].append(Host(ip_addr, user="root"))
        else:
            roles[node_role] = [Host(ip_addr, user="root")]

    if ctx.use_httpd:
        ctx.httpd.stop()
    return roles




import os
import os.path as op
import time
from string import Template
import click
import subprocess
import socket

from ..flavour import Flavour
from ..actions import (
    read_compose_info,
    realpath_from_store,
    generate_deployment_info,
    generate_deploy_info_b64,
    generate_kexec_scripts,
    launch_ssh_kexec,
    wait_ssh_ports,
    ssh_connect,
)
from ..driver.machine import Machine

# from ..driver.logger import rootlog


KADEPOY_ARCH = {
    "x86_64-linux": "x86_64",
    "powerpc64le-linux": "ppc64le",
    "aarch64-linux": "aarch64",
}

KADEPOY_ENV_DESC = """
      name: $image_name
      version: 1
      description: NixOS
      author: $author
      visibility: shared
      destructive: false
      os: linux
      arch: $system
      image:
        file: $file_image_url
        kind: tar
        compression: xz
      boot:
        kernel: /boot/bzImage
        initrd: /boot/initrd
        kernel_params: $kernel_params
      filesystem: ext4
      partition_type: 131
      multipart: false
"""


def generate_kadeploy_envfile(ctx, deploy=None, kernel_params_opts=""):
    if not ctx.compose_info:
        read_compose_info(ctx)

    base_path = op.join(
        ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
    )
    os.makedirs(base_path, mode=0o700, exist_ok=True)
    kaenv_path = op.join(base_path, "nixos.yaml")

    if not deploy:
        if ctx.use_httpd:
            base_url = f"http://{ctx.httpd.ip}:{ctx.httpd.port}"
            deploy = f"{base_url}/deploy/{ctx.composition_flavour_prefix}.json"
        else:
            generate_deploy_info_b64(ctx)
            deploy = ctx.deployment_info_b64

    user = os.environ["USER"]
    system = ctx.compositions_info["system"]
    with open(kaenv_path, "w") as kaenv_file:
        t = Template(KADEPOY_ENV_DESC)
        kaenv = t.substitute(
            image_name="NixOS",
            author=user,
            system=KADEPOY_ARCH[system],
            file_image_url=f"http://public.grenoble.grid5000.fr/~{user}/nixos.tar.xz",
            kernel_params=f"boot.shell_on_fail console=tty0 console=ttyS0,115200 deploy={deploy} {kernel_params_opts}",
        )
        kaenv_file.write(kaenv)


class G5kKexecBasedFlavour(Flavour):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate_deployment_info(self):
        generate_deployment_info(self.ctx)

    def generate_kexec_scripts(self):
        generate_kexec_scripts(self.ctx)

    def launch(self):
        launch_ssh_kexec(self.ctx)
        time.sleep(10)
        wait_ssh_ports(self.ctx)

    def driver_initialize(self, tmp_dir):
        self.tmp_dir = tmp_dir
        ctx = self.ctx

        if ctx.no_start:  #
            deployment_nodes = self.ctx.deployment_info["deployment"]
            for ip, node in deployment_nodes.items():
                self.machines.append(
                    Machine(
                        self.ctx,
                        ip=ip,
                        tmp_dir=tmp_dir,
                        start_command="",
                        keep_vm_state=False,
                        name=node["host"],
                    )
                )

            for machine in self.machines:
                if not machine.connected:
                    self.start(machine)
                machine.connected = True
            return

    def start(self, machine):
        if not self.ctx.no_start:
            print("Not Yet Implemented")
            exit(1)
        else:
            machine.start_process_shell(
                [
                    "ssh",
                    "-t",
                    "-o",
                    "StrictHostKeyChecking=no",
                    "-l",
                    "root",
                    machine.ip,
                ]
            )

    def ext_connect(self, user, node, execute=True):
        return ssh_connect(self.ctx, user, node, execute)


class G5kNfsStoreFlavour(G5kKexecBasedFlavour):
    def __init__(self, ctx):
        super().__init__(ctx)

        self.name = "g5k-nfs-store"

    def generate_kexec_scripts(self):
        def store_path():
            for prefix_store_path in self.ctx.alternative_stores + ["/nix"]:
                store_path = f"{prefix_store_path}/store"
                if op.exists(store_path):
                    return store_path
            raise "Store Path Not Found"

        if "NFS_STORE" in os.environ:
            kernel_params = f"nfs_store={os.environ['NFS_STORE']}"
        else:
            nfs = socket.getfqdn("nfs")
            store_path = store_path()
            kernel_params = f"nfs_store={nfs}:/export{store_path}"
        self.ctx.vlog(f" kernel_params: {kernel_params}")

        generate_kexec_scripts(self.ctx, flavour_kernel_params=kernel_params)


class G5kRamdiskFlavour(G5kKexecBasedFlavour):
    def __init__(self, ctx):
        super().__init__(ctx)

        self.name = "g5k-ramdisk"


class G5KImageFlavour(Flavour):
    def __init__(self, ctx):
        super().__init__(ctx)

        self.name = "g5k-image"

    def generate_deployment_info(self):
        generate_deployment_info(self.ctx)

    def launch(self, machine_file=None):
        generate_kadeploy_envfile(self.ctx)
        image_path = realpath_from_store(
            self.ctx, self.ctx.deployment_info["all"]["image"]
        )
        cmd_copy_image = f'cp {image_path} ~{os.environ["USER"]}/public/nixos.tar.xz && chmod 644 ~{os.environ["USER"]}/public/nixos.tar.xz'
        if machine_file or click.confirm(
            f'Do you want to copy image to ~{os.environ["USER"]}/public/nixos.tar.xz ?'
        ):
            try:
                subprocess.call(cmd_copy_image, shell=True)
            except Exception as ex:
                raise click.ClickException(f"Failed to copy image: {ex}")
        else:
            print(f"You can copy image with: {cmd_copy_image}")
        base_path = op.join(
            self.ctx.envdir,
            f"artifact/{self.ctx.composition_name}/{self.ctx.flavour.name}",
        )
        if machine_file:
            cmd_kadeploy = (
                f'kadeploy3 -a {op.join(base_path, "nixos.yaml")} -f {machine_file} -k'
            )
        else:
            cmd_kadeploy = (
                f'kadeploy3 -a {op.join(base_path, "nixos.yaml")} -f $OAR_NODEFILE -k'
            )

        if machine_file or click.confirm(
            "Do you want to kadeploy nixos.tar.xz image on nodes from $OAR_NODEFILE"
        ):
            try:
                subprocess.call(cmd_kadeploy, shell=True)
            except Exception as ex:
                raise click.ClickException(f"Failed to execute kadeploy command: {ex}")
        else:
            print(f"You can kadeploy image with: {cmd_kadeploy}")

    def start(self, machine):
        if not self.ctx.no_start:
            print("Not Yet Implemented")
            exit(1)
        else:
            machine.start_process_shell(
                [
                    "ssh",
                    "-t",
                    "-o",
                    "StrictHostKeyChecking=no",
                    "-l",
                    "root",
                    machine.ip,
                ]
            )

    def ext_connect(self, user, node, execute=True):
        return ssh_connect(self.ctx, user, node, execute)



import click
import json
import re
import sys
import ptpython.repl

from ..context import pass_context
from ..actions import read_deployment_info, realpath_from_store
from ..flavours import get_flavour_by_name

from ..driver.driver import Driver


@click.command("driver")
@click.option("-l", "--user", default="root")
@click.option(
    "-d",
    "--deployment-file",
    help="Deployment file, take the latest created in deploy directory by default",
)
@click.option(
    "-f",
    "--flavour",
    help="flavour, by default it's extracted from deployment file name",
)
@click.option(
    "-t", "--test-script", is_flag=True, help="execute the 'embedded' testscript",
)
@click.argument("test-script-file", required=False)
@pass_context
# TODO @on_finished(lambda ctx: ctx.state.dump())
# TODO @on_started(lambda ctx: ctx.assert_valid_env())
def cli(ctx, user, deployment_file, flavour, test_script_file, test_script):
    """Start driver to intearct with deployed environment."""
    read_deployment_info(ctx, deployment_file)

    # determine flavour name
    if not flavour:
        match = re.match(r"^.*::(.+)\..*$", ctx.deployment_filename)
        if match:
            flavour = match.group(1)
        else:
            raise click.ClickException(
                "Cannot determined flavour, one must by provided (option fix deployment file name or use --flavour option"
            )

    ctx.flavour = get_flavour_by_name(flavour)(ctx)

    ctx.external_connect = True
    ctx.no_start = True

    # TODO add test to ssh port
    if test_script:
        if "test_script" in ctx.deployment_info:
            test_script_file = ctx.deployment_info["test_script"]

        elif "compositions_info_path" in ctx.deployment_info:
            compositions_info_file = realpath_from_store(
                ctx, ctx.deployment_info["compositions_info_path"]
            )
            with open(compositions_info_file, "r") as f:
                compositions_info = json.load(f)

            selected_composition = ctx.deployment_info["composition"]
            test_script_file = compositions_info[selected_composition]["test_script"]

        test_script_file = realpath_from_store(ctx, test_script_file)

    test_script_str = ""
    if test_script_file:
        test_script = True
        with open(test_script_file, "r") as f:
            test_script_str = f.read()

    if test_script and test_script_str == "":
        ctx.elog(f"test script ({test_script_file}) is empty")
        sys.exit(1)

    with Driver(
        # args.start_scripts, args.vlans, args.testscript.read_text(), args.keep_vm_state
        ctx,
        [],
        [],
        test_script_str,
        False,
    ) as driver:
        if test_script:
            try:
                driver.test_script()
            except Exception as e:
                ctx.elog(e)
                sys.exit(1)
        else:
            ptpython.repl.embed(driver.test_symbols(), {})



from contextlib import _GeneratorContextManager
from pathlib import Path
from queue import Queue
from typing import Any, Callable, Dict, List, Optional, Tuple
import base64
import io
import os
import queue
import re
import shlex
import shutil
import socket
import subprocess
import sys
import tempfile
import threading
import time

from .logger import rootlog
from ..flavours import use_flavour_method_if_any

CHAR_TO_KEY = {
    "A": "shift-a",
    "N": "shift-n",
    "-": "0x0C",
    "_": "shift-0x0C",
    "B": "shift-b",
    "O": "shift-o",
    "=": "0x0D",
    "+": "shift-0x0D",
    "C": "shift-c",
    "P": "shift-p",
    "[": "0x1A",
    "{": "shift-0x1A",
    "D": "shift-d",
    "Q": "shift-q",
    "]": "0x1B",
    "}": "shift-0x1B",
    "E": "shift-e",
    "R": "shift-r",
    ";": "0x27",
    ":": "shift-0x27",
    "F": "shift-f",
    "S": "shift-s",
    "'": "0x28",
    '"': "shift-0x28",
    "G": "shift-g",
    "T": "shift-t",
    "`": "0x29",
    "~": "shift-0x29",
    "H": "shift-h",
    "U": "shift-u",
    "\\": "0x2B",
    "|": "shift-0x2B",
    "I": "shift-i",
    "V": "shift-v",
    ",": "0x33",
    "<": "shift-0x33",
    "J": "shift-j",
    "W": "shift-w",
    ".": "0x34",
    ">": "shift-0x34",
    "K": "shift-k",
    "X": "shift-x",
    "/": "0x35",
    "?": "shift-0x35",
    "L": "shift-l",
    "Y": "shift-y",
    " ": "spc",
    "M": "shift-m",
    "Z": "shift-z",
    "\n": "ret",
    "!": "shift-0x02",
    "@": "shift-0x03",
    "#": "shift-0x04",
    "$": "shift-0x05",
    "%": "shift-0x06",
    "^": "shift-0x07",
    "&": "shift-0x08",
    "*": "shift-0x09",
    "(": "shift-0x0A",
    ")": "shift-0x0B",
}


def make_command(args: list) -> str:
    return " ".join(map(shlex.quote, (map(str, args))))


def retry(fn: Callable, timeout: int = 900) -> None:
    """Call the given function repeatedly, with 1 second intervals,
    until it returns True or a timeout is reached.
    """

    for _ in range(timeout):
        if fn(False):
            return
        time.sleep(1)

    if not fn(True):
        raise Exception(f"action timed out after {timeout} seconds")


class StartCommand:
    pass


class VmStartCommand(StartCommand):
    """The Base Start Command knows how to append the necesary
    runtime qemu options as determined by a particular test driver
    run. Any such start command is expected to happily receive and
    append additional qemu args.
    """

    _cmd: str = ""
    qemu_opts: str = ""

    def __init__(self):
        pass

    def cmd(
        self,
        monitor_socket_path: Path,
        shell_socket_path: Path,
        allow_reboot: bool = False,  # TODO: unused, legacy?
    ) -> str:
        display_opts = ""
        display_available = any(x in os.environ for x in ["DISPLAY", "WAYLAND_DISPLAY"])
        if not display_available:
            display_opts += " -nographic"

        # qemu options
        qemu_opts = ""
        qemu_opts += (
            ""
            if allow_reboot
            else " -no-reboot"
            " -device virtio-serial"
            " -device virtconsole,chardev=shell"
            " -device virtio-rng-pci"
            " -serial stdio"
            f" -monitor unix:{monitor_socket_path}"
            f" -chardev socket,id=shell,path={shell_socket_path}"
        )
        # TODO: qemu script already catpures this env variable, legacy?
        qemu_opts += " " + os.environ.get("QEMU_OPTS", "")

        self.qemu_opts = qemu_opts

        return f"{self._cmd}"

    def build_environment(self, state_dir: Path, shared_dir: Path,) -> dict:
        # We make a copy to not update the current environment
        env = dict(os.environ)
        env.update(
            {
                "TMPDIR": str(state_dir),
                "SHARED_DIR": str(shared_dir),
                "USE_TMPDIR": "1",
                "QEMU_OPTS": self.qemu_opts,
                "VM_ID": str(self.vm_id),
                "QEMU_VDE_SOCKET": str(self.flavour.vlan.socket_dir),
            }
        )
        return env

    def run(
        self,
        state_dir: Path,
        shared_dir: Path,
        monitor_socket_path: Path,
        shell_socket_path: Path,
    ) -> subprocess.Popen:
        return subprocess.Popen(
            self.cmd(monitor_socket_path, shell_socket_path),
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            cwd=state_dir,
            env=self.build_environment(state_dir, shared_dir),
        )


class StartScript(VmStartCommand):
    def __init__(self, script: str, vm_id: str, flavour):
        super().__init__()
        self._cmd = script
        self.vm_id = vm_id
        self.flavour = flavour


class NixStartScript(VmStartCommand):
    """A start script from nixos/modules/virtualiation/qemu-vm.nix
    that also satisfies the requirement of the BaseStartCommand.
    These Nix commands have the particular charactersitic that the
    machine name can be extracted out of them via a regex match.
    (Admittedly a _very_ implicit contract, evtl. TODO fix)
    """

    def __init__(self, script: str, vm_id: str):
        self._cmd = script

    @property
    def machine_name(self) -> str:
        match = re.search("run-(.+)-vm$", self._cmd)
        name = "machine"
        if match:
            name = match.group(1)
        return name


class Machine:
    """A handle to the machine with this name, that also knows how to manage
    the machine lifecycle with the help of a start script / command."""

    name: str
    tmp_dir: Path
    shared_dir: Path
    state_dir: Path
    monitor_path: Path
    shell_path: Path

    start_command: StartCommand
    keep_vm_state: bool
    allow_reboot: bool

    process: Optional[subprocess.Popen]
    pid: Optional[int]
    monitor: Optional[socket.socket]
    shell: Optional[socket.socket]
    serial_thread: Optional[threading.Thread]
    process_shell: Optional[subprocess.Popen]

    booted: bool
    connected: bool
    # Store last serial console lines for use
    # of wait_for_console_text
    last_lines: Queue = Queue()
    ctx = None

    def __repr__(self) -> str:
        return f"<Machine '{self.name}'>"

    def __init__(
        self,
        ctx,
        tmp_dir: Path,
        start_command: StartCommand,
        name: str = "machine",
        ip: str = "",
        keep_vm_state: bool = False,
        allow_reboot: bool = False,
        vm_id: str = "",
        init: str = "",
    ) -> None:
        self.ctx = ctx
        self.tmp_dir = tmp_dir
        self.keep_vm_state = keep_vm_state
        self.allow_reboot = allow_reboot
        self.name = name
        self.start_command = start_command
        self.ip = ip
        self.vm_id = vm_id
        self.init = init

        # set up directories (TODO only for vm ???)
        self.shared_dir = self.tmp_dir / "shared-xchg"
        self.shared_dir.mkdir(mode=0o700, exist_ok=True)

        self.state_dir = self.tmp_dir / f"vm-state-{self.name}"
        self.monitor_path = self.state_dir / "monitor"
        self.shell_path = self.state_dir / "shell"
        if (not self.keep_vm_state) and self.state_dir.exists():
            self.cleanup_statedir()
        self.state_dir.mkdir(mode=0o700, exist_ok=True)

        self.process = None
        self.pid = None
        self.monitor = None
        self.shell = None
        self.serial_thread = None

        self.booted = False
        self.connected = False

    def is_up(self) -> bool:
        return self.booted and self.connected

    def log(self, msg: str) -> None:
        rootlog.log(msg, {"machine": self.name})

    def log_serial(self, msg: str) -> None:
        rootlog.log_serial(msg, self.name)

    def nested(self, msg: str, attrs: Dict[str, str] = {}) -> _GeneratorContextManager:
        my_attrs = {"machine": self.name}
        my_attrs.update(attrs)
        return rootlog.nested(msg, my_attrs)

    def wait_for_monitor_prompt(self) -> str:
        with self.nested("waiting for monitor prompt"):
            assert self.monitor is not None
            answer = ""
            while True:
                undecoded_answer = self.monitor.recv(1024)
                if not undecoded_answer:
                    break
                answer += undecoded_answer.decode()
                if answer.endswith("(qemu) "):
                    break
            return answer

    def send_monitor_command(self, command: str) -> str:
        with self.nested("sending monitor command: {}".format(command)):
            message = ("{}\n".format(command)).encode()
            assert self.monitor is not None
            self.monitor.send(message)
            return self.wait_for_monitor_prompt()

    def wait_for_unit(self, unit: str, user: Optional[str] = None) -> None:
        """Wait for a systemd unit to get into "active" state.
        Throws exceptions on "failed" and "inactive" states as well as
        after timing out.
        """

        def check_active(_: Any) -> bool:
            info = self.get_unit_info(unit, user)
            state = info["ActiveState"]
            if state == "failed":
                raise Exception('unit "{}" reached state "{}"'.format(unit, state))

            if state == "inactive":
                status, jobs = self.systemctl("list-jobs --full 2>&1", user)
                if "No jobs" in jobs:
                    info = self.get_unit_info(unit, user)
                    if info["ActiveState"] == state:
                        raise Exception(
                            (
                                'unit "{}" is inactive and there ' "are no pending jobs"
                            ).format(unit)
                        )

            return state == "active"

        with self.nested(
            "waiting for unit {}{}".format(
                unit, f" with user {user}" if user is not None else ""
            )
        ):
            retry(check_active)

    def get_unit_info(self, unit: str, user: Optional[str] = None) -> Dict[str, str]:
        status, lines = self.systemctl('--no-pager show "{}"'.format(unit), user)
        if status != 0:
            raise Exception(
                'retrieving systemctl info for unit "{}" {} failed with exit code {}'.format(
                    unit, "" if user is None else 'under user "{}"'.format(user), status
                )
            )

        line_pattern = re.compile(r"^([^=]+)=(.*)$")

        def tuple_from_line(line: str) -> Tuple[str, str]:
            match = line_pattern.match(line)
            assert match is not None
            return match[1], match[2]

        return dict(
            tuple_from_line(line)
            for line in lines.split("\n")
            if line_pattern.match(line)
        )

    def systemctl(self, q: str, user: Optional[str] = None) -> Tuple[int, str]:
        if user is not None:
            q = q.replace("'", "\\'")
            return self.execute(
                (
                    "su -l {} --shell /bin/sh -c "
                    "$'XDG_RUNTIME_DIR=/run/user/`id -u` "
                    "systemctl --user {}'"
                ).format(user, q)
            )
        return self.execute("systemctl {}".format(q))

    def require_unit_state(self, unit: str, require_state: str = "active") -> None:
        with self.nested(
            "checking if unit ‘{}’ has reached state '{}'".format(unit, require_state)
        ):
            info = self.get_unit_info(unit)
            state = info["ActiveState"]
            if state != require_state:
                raise Exception(
                    "Expected unit ‘{}’ to to be in state ".format(unit)
                    + "'{}' but it is in state ‘{}’".format(require_state, state)
                )

    def _next_newline_closed_block_from_shell(self) -> str:
        assert self.shell
        output_buffer = []
        while True:
            # This receives up to 4096 bytes from the socket
            chunk = self.shell.recv(4096)
            if not chunk:
                # Probably a broken pipe, return the output we have
                break

            decoded = chunk.decode()
            output_buffer += [decoded]
            if decoded[-1] == "\n":
                break
        return "".join(output_buffer)

    @use_flavour_method_if_any
    def execute(
        self, command: str, check_return: bool = True, timeout: Optional[int] = 900
    ) -> Tuple[int, str]:

        if self.ctx.external_connect:
            return self.execute_process_shell(command, check_return, timeout)

        self.connect()

        if timeout is not None:
            command = "timeout {} sh -c {}".format(timeout, shlex.quote(command))

        out_command = f"( set -euo pipefail; {command} ) | (base64 --wrap 0; echo)\n"
        assert self.shell
        self.shell.send(out_command.encode())

        # Get the output
        output = base64.b64decode(self._next_newline_closed_block_from_shell())

        if not check_return:
            return (-1, output.decode())

        # Get the return code
        self.shell.send("echo ${PIPESTATUS[0]}\n".encode())
        rc = int(self._next_newline_closed_block_from_shell().strip())

        return (rc, output.decode())

    @use_flavour_method_if_any
    def shell_interact(self) -> None:
        """Allows you to interact with the guest shell
        Should only be used during test development, not in the production test."""

        self.connect()
        self.log("Terminal is ready (there is no prompt):")

        assert self.shell
        subprocess.run(
            ["socat", "READLINE", f"FD:{self.shell.fileno()}"],
            pass_fds=[self.shell.fileno()],
        )

    def succeed(self, *commands: str, timeout: Optional[int] = None) -> str:
        """Execute each command and check that it succeeds."""
        output = ""
        for command in commands:
            with self.nested("must succeed: {}".format(command)):
                (status, out) = self.execute(command, timeout=timeout)
                if status != 0:
                    self.log("output: {}".format(out))
                    raise Exception(
                        "command `{}` failed (exit code {})".format(command, status)
                    )
                output += out
        return output

    def fail(self, *commands: str, timeout: Optional[int] = None) -> str:
        """Execute each command and check that it fails."""
        output = ""
        for command in commands:
            with self.nested("must fail: {}".format(command)):
                (status, out) = self.execute(command, timeout=timeout)
                if status == 0:
                    raise Exception(
                        "command `{}` unexpectedly succeeded".format(command)
                    )
                output += out
        return output

    def wait_until_succeeds(self, command: str, timeout: int = 900) -> str:
        """Wait until a command returns success and return its output.
        Throws an exception on timeout.
        """
        output = ""

        def check_success(_: Any) -> bool:
            nonlocal output
            status, output = self.execute(command, timeout=timeout)
            return status == 0

        with self.nested("waiting for success: {}".format(command)):
            retry(check_success, timeout)
            return output

    def wait_until_fails(self, command: str, timeout: int = 900) -> str:
        """Wait until a command returns failure.
        Throws an exception on timeout.
        """
        output = ""

        def check_failure(_: Any) -> bool:
            nonlocal output
            status, output = self.execute(command, timeout=timeout)
            return status != 0

        with self.nested("waiting for failure: {}".format(command)):
            retry(check_failure)
            return output

    def wait_for_shutdown(self) -> None:
        if not self.booted:
            return

        with self.nested("waiting for the VM to power off"):
            sys.stdout.flush()
            assert self.process
            self.process.wait()

            self.pid = None
            self.booted = False
            self.connected = False

    def get_tty_text(self, tty: str) -> str:
        status, output = self.execute(
            "fold -w$(stty -F /dev/tty{0} size | "
            "awk '{{print $2}}') /dev/vcs{0}".format(tty)
        )
        return output

    def wait_until_tty_matches(self, tty: str, regexp: str) -> None:
        """Wait until the visible output on the chosen TTY matches regular
        expression. Throws an exception on timeout.
        """
        matcher = re.compile(regexp)

        def tty_matches(last: bool) -> bool:
            text = self.get_tty_text(tty)
            if last:
                self.log(
                    f"Last chance to match /{regexp}/ on TTY{tty}, "
                    f"which currently contains: {text}"
                )
            return len(matcher.findall(text)) > 0

        with self.nested("waiting for {} to appear on tty {}".format(regexp, tty)):
            retry(tty_matches)

    def send_chars(self, chars: List[str]) -> None:
        with self.nested("sending keys ‘{}‘".format(chars)):
            for char in chars:
                self.send_key(char)

    def wait_for_file(self, filename: str) -> None:
        """Waits until the file exists in machine's file system."""

        def check_file(_: Any) -> bool:
            status, _ = self.execute("test -e {}".format(filename))
            return status == 0

        with self.nested("waiting for file ‘{}‘".format(filename)):
            retry(check_file)

    def wait_for_open_port(self, port: int) -> None:
        def port_is_open(_: Any) -> bool:
            status, _ = self.execute("nc -z localhost {}".format(port))
            return status == 0

        with self.nested("waiting for TCP port {}".format(port)):
            retry(port_is_open)

    def wait_for_closed_port(self, port: int) -> None:
        def port_is_closed(_: Any) -> bool:
            status, _ = self.execute("nc -z localhost {}".format(port))
            return status != 0

        with self.nested("waiting for TCP port {} to be closed"):
            retry(port_is_closed)

    def start_job(self, jobname: str, user: Optional[str] = None) -> Tuple[int, str]:
        return self.systemctl("start {}".format(jobname), user)

    def stop_job(self, jobname: str, user: Optional[str] = None) -> Tuple[int, str]:
        return self.systemctl("stop {}".format(jobname), user)

    def wait_for_job(self, jobname: str) -> None:
        self.wait_for_unit(jobname)

    @use_flavour_method_if_any
    def connect(self) -> None:
        if self.connected:
            return

        with self.nested("waiting for the VM to finish booting"):
            self.start()

            assert self.shell

            tic = time.time()
            self.shell.recv(1024)
            # TODO: Timeout
            toc = time.time()

            self.log("connected to guest root shell")
            self.log("(connecting took {:.2f} seconds)".format(toc - tic))
            self.connected = True

    def copy_from_host_via_shell(self, source: str, target: str) -> None:
        """Copy a file from the host into the guest by piping it over the
        shell into the destination file. Works without host-guest shared folder.
        Prefer copy_from_host for whenever possible.
        """
        with open(source, "rb") as fh:
            content_b64 = base64.b64encode(fh.read()).decode()
            self.succeed(
                f"mkdir -p $(dirname {target})",
                f"echo -n {content_b64} | base64 -d > {target}",
            )

    def copy_from_host(self, source: str, target: str) -> None:
        """Copy a file from the host into the guest via the `shared_dir` shared
        among all the VMs (using a temporary directory).
        """
        host_src = Path(source)
        vm_target = Path(target)
        with tempfile.TemporaryDirectory(dir=self.shared_dir) as shared_td:
            shared_temp = Path(shared_td)
            host_intermediate = shared_temp / host_src.name
            vm_shared_temp = Path("/tmp/shared") / shared_temp.name
            vm_intermediate = vm_shared_temp / host_src.name

            self.succeed(make_command(["mkdir", "-p", vm_shared_temp]))
            if host_src.is_dir():
                shutil.copytree(host_src, host_intermediate)
            else:
                shutil.copy(host_src, host_intermediate)
            self.succeed(make_command(["mkdir", "-p", vm_target.parent]))
            self.succeed(make_command(["cp", "-r", vm_intermediate, vm_target]))

    def copy_from_vm(self, source: str, target_dir: str = "") -> None:
        """Copy a file from the VM (specified by an in-VM source path) to a path
        relative to `$out`. The file is copied via the `shared_dir` shared among
        all the VMs (using a temporary directory).
        """
        # Compute the source, target, and intermediate shared file names
        out_dir = Path(os.environ.get("out", os.getcwd()))
        vm_src = Path(source)
        with tempfile.TemporaryDirectory(dir=self.shared_dir) as shared_td:
            shared_temp = Path(shared_td)
            vm_shared_temp = Path("/tmp/shared") / shared_temp.name
            vm_intermediate = vm_shared_temp / vm_src.name
            intermediate = shared_temp / vm_src.name
            # Copy the file to the shared directory inside VM
            self.succeed(make_command(["mkdir", "-p", vm_shared_temp]))
            self.succeed(make_command(["cp", "-r", vm_src, vm_intermediate]))
            abs_target = out_dir / target_dir / vm_src.name
            abs_target.parent.mkdir(exist_ok=True, parents=True)
            # Copy the file from the shared directory outside VM
            if intermediate.is_dir():
                shutil.copytree(intermediate, abs_target)
            else:
                shutil.copy(intermediate, abs_target)

    def dump_tty_contents(self, tty: str) -> None:
        """Debugging: Dump the contents of the TTY<n>"""
        self.execute("fold -w 80 /dev/vcs{} | systemd-cat".format(tty))

    def wait_for_console_text(self, regex: str) -> None:
        with self.nested("waiting for {} to appear on console".format(regex)):
            # Buffer the console output, this is needed
            # to match multiline regexes.
            console = io.StringIO()
            while True:
                try:
                    console.write(self.last_lines.get())
                except queue.Empty:
                    self.sleep(1)
                    continue
                console.seek(0)
                matches = re.search(regex, console.read())
                if matches is not None:
                    return

    def send_key(self, key: str) -> None:
        key = CHAR_TO_KEY.get(key, key)
        self.send_monitor_command("sendkey {}".format(key))
        time.sleep(0.01)

    def start(self) -> None:
        self.ctx.flavour.start(self)

    def _start_vm(self) -> None:
        """Should work for vm-ramdisk (tested) and other qemu based (not tested)"""
        if self.booted:
            return

        self.log("starting vm")

        def clear(path: Path) -> Path:
            if path.exists():
                path.unlink()
            return path

        def create_socket(path: Path) -> socket.socket:
            s = socket.socket(family=socket.AF_UNIX, type=socket.SOCK_STREAM)
            s.bind(str(path))
            s.listen(1)
            return s

        monitor_socket = create_socket(clear(self.monitor_path))
        shell_socket = create_socket(clear(self.shell_path))
        self.process = self.start_command.run(
            self.state_dir, self.shared_dir, self.monitor_path, self.shell_path,
        )
        self.monitor, _ = monitor_socket.accept()
        self.shell, _ = shell_socket.accept()

        # Store last serial console lines for use
        # of wait_for_console_text
        self.last_lines: Queue = Queue()

        def process_serial_output() -> None:
            assert self.process
            assert self.process.stdout
            for _line in self.process.stdout:
                # Ignore undecodable bytes that may occur in boot menus
                line = _line.decode(errors="ignore").replace("\r", "").rstrip()
                self.last_lines.put(line)
                self.log_serial(line)

        self.serial_thread = threading.Thread(target=process_serial_output)
        # self.serial_thread.daemon = True
        self.serial_thread.start()

        self.wait_for_monitor_prompt()

        self.pid = self.process.pid
        self.booted = True

        self.log("QEMU running (pid {})".format(self.pid))

    def cleanup_statedir(self) -> None:
        if self.ctx.flavour in [
            "vm",
            "vm-bridged",
            "nixos-test",
            "nixos-test-driver",
            "nixos-test-ssh",
        ]:
            shutil.rmtree(self.state_dir)
            rootlog.log(f"deleting VM state directory {self.state_dir}")
            rootlog.log("if you want to keep the VM state, pass --keep-vm-state")

    def shutdown(self) -> None:
        if not self.booted:
            return

        assert self.shell
        self.shell.send("poweroff\n".encode())
        self.wait_for_shutdown()

    def crash(self) -> None:
        if not self.booted:
            return

        self.log("forced crash")
        self.send_monitor_command("quit")
        self.wait_for_shutdown()

    def wait_for_x(self) -> None:
        """Wait until it is possible to connect to the X server.  Note that
        testing the existence of /tmp/.X11-unix/X0 is insufficient.
        """

        def check_x(_: Any) -> bool:
            cmd = (
                "journalctl -b SYSLOG_IDENTIFIER=systemd | "
                + 'grep "Reached target Current graphical"'
            )
            status, _ = self.execute(cmd)
            if status != 0:
                return False
            status, _ = self.execute("[ -e /tmp/.X11-unix/X0 ]")
            return status == 0

        with self.nested("waiting for the X11 server"):
            retry(check_x)

    def get_window_names(self) -> List[str]:
        return self.succeed(
            r"xwininfo -root -tree | sed 's/.*0x[0-9a-f]* \"\([^\"]*\)\".*/\1/; t; d'"
        ).splitlines()

    def wait_for_window(self, regexp: str) -> None:
        pattern = re.compile(regexp)

        def window_is_visible(last_try: bool) -> bool:
            names = self.get_window_names()
            if last_try:
                self.log(
                    "Last chance to match {} on the window list,".format(regexp)
                    + " which currently contains: "
                    + ", ".join(names)
                )
            return any(pattern.search(name) for name in names)

        with self.nested("waiting for a window to appear"):
            retry(window_is_visible)

    def sleep(self, secs: int) -> None:
        # We want to sleep in *guest* time, not *host* time.
        self.succeed(f"sleep {secs}")

    def forward_port(self, host_port: int = 8080, guest_port: int = 80) -> None:
        """Forward a TCP port on the host to a TCP port on the guest.
        Useful during interactive testing.
        """
        self.send_monitor_command(
            "hostfwd_add tcp::{}-:{}".format(host_port, guest_port)
        )

    def block(self) -> None:
        """Make the machine unreachable by shutting down eth1 (the multicast
        interface used to talk to the other VMs).  We keep eth0 up so that
        the test driver can continue to talk to the machine.
        """
        self.send_monitor_command("set_link virtio-net-pci.1 off")

    def unblock(self) -> None:
        """Make the machine reachable."""
        self.send_monitor_command("set_link virtio-net-pci.1 on")

    @use_flavour_method_if_any
    def release(self) -> None:

        if self.pid is None:
            return

        rootlog.info(f"kill machine (pid {self.pid})")

        assert self.process
        assert self.shell
        assert self.monitor
        assert self.serial_thread

        self.process.terminate()
        self.shell.close()
        self.monitor.close()
        self.serial_thread.join()

    def start_process_shell(self, args):
        # command examples:
        # ['docker-compose', '-f', '/home/auguste/work/tests/21-22/nxc-test/nxc/deploy/docker_compose/docker-compose.json', 'exec', '-T', ']

        # ['ssh', '-t', '-o', 'StrictHostKeyChecking=no', '-l', 'root', '10.0.2.16']
        self.process_shell = subprocess.Popen(
            args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        )

    def execute_process_shell(
        self, command: str, check_return: bool = True, timeout: Optional[int] = 900,
    ) -> Tuple[int, str]:

        self.connect()

        process_shell = self.process_shell

        try:
            (stdout, _stderr) = process_shell.communicate(
                command.encode(), timeout=timeout
            )
        except subprocess.TimeoutExpired:
            process_shell.kill()
            return (-1, "")

        status_code = process_shell.returncode
        self.restart_process_shell()

        if not check_return:
            return (-1, stdout.decode())

        return (status_code, stdout.decode())

    def restart_process_shell(self):
        self.start()



import click
from ..context import pass_context
from ..g5k import key_sleep_script
from ..actions import install_nix_static


def print_helper(ctx, options):
    for option in options:
        if (option == "g5k_script") or (option == "g5k-script"):
            click.echo(key_sleep_script)
        elif option == "install-nix":
            install_nix_static()
        else:
            ctx.elog(f"Helper: {option} does not exist")


def print_helper_list(helper_options):
    print("g5k-script: print path to g5k_key_sleep_script Grid'5000 script")
    print("install-nix: install the nix command in ~/.local/bin")


@click.command("helper")
@click.option(
    "-l",
    "--list",
    # "--list-helpers",
    is_flag=True,
    help="List of available helpers",
)
@pass_context
@click.argument("options", nargs=-1)
def cli(ctx, options, list):
    """Specific and contextual helper information (e.g. g5k_script path for Grid'5000)
    Warning: Experimental command, may be removed in the future or change without backward compatibility care."""
    if options:
        print_helper(ctx, options)

    if list:
        print_helper_list(ctx)



import os
import sys
import time

import json
import yaml

from io import open
from functools import update_wrapper

import click

from .platform import Grid5000Platform
from .default_role import get_nxc_loader

# from .state import State

CONTEXT_SETTINGS = dict(
    auto_envvar_prefix="nixos_compose", help_option_names=["-h", "--help"]
)


def reraise(tp, value, tb=None):
    if value.__traceback__ is not tb:
        raise value.with_traceback(tb)
    raise value


class Context(object):
    def __init__(self):
        self.t0 = time.time()
        self.nxc_file = None
        self.nxc = None
        self.current_dir = os.getcwd()  # TOREMOVE ?
        self.verbose = False
        self.workdir = self.current_dir  # TOREMOVE ?
        self.debug = False
        self.prefix = "nxc"
        self.flavour = None
        self.composition_name = None
        self.composition_flavour_prefix = None  # REMOVE _prefix ???
        self.compose_info_file = None  # change to compositions_file ?
        self.compose_info = None  # change to compositioin of merge w/ compositionS ?
        self.multiple_compositions = False
        self.compositions_info = None  # Rename to compositions_flavour ????
        self.deployment_filename: str = ""
        self.deployment_info = None  # change to deployment ?
        self.deployment_info_b64 = ""  # change to depolyment_b64 ?
        self.ip_addresses = []
        self.host2ip_address = {}
        self.ssh = ""
        self.sudo = ""
        self.push_path = None
        self.interactive = False
        self.execute_test_script = False
        self.platform = None
        self.use_httpd = False
        self.httpd = None
        self.forward_ssh_port = False
        self.alternative_stores = [
            f"{os.environ['HOME']}/.local/share/nix/root/nix",
            f"{os.environ['HOME']}/.nix",
        ]
        self.roles_quantities = {}
        self.setup = None
        self.sigwait = None
        self.all_started: bool = False
        self.no_start: bool = False
        self.external_connect: bool = False

    def init_workdir(self, env_name, env_id):
        with open(self.env_name_file, "w+") as fd:
            fd.write(env_name + "\n")
        if not os.path.exists(self.env_id_file):
            with open(self.env_id_file, "w+") as fd:
                fd.write(env_id + "\n")

    #    @property
    #    def state(self):
    #        if not hasattr(self, "_state"):
    #            self._state = State(self, state_file=self.state_file)
    #        return self._state
    #
    #    def update(self):
    #        self.state_file = op.join(self.envdir, "state.json")
    #        if "platform" in self.state:
    #            if self.state["platform"] == "Grid5000":
    #                self.platform = Grid5000Platform(self)

    def assert_valid_env(self):
        if not os.path.isdir(self.envdir):
            raise click.ClickException(
                "Missing nixos composition environment directory."
                " Run `nxc init` to create"
                " a new composition environment "
            )

    def log(self, msg, *args, **kwargs):
        """Logs a message to stdout."""
        if args:
            msg %= args
        kwargs.setdefault("file", sys.stdout)
        click.echo(msg, **kwargs)

    def wlog(self, msg, *args):
        """Logs a warning message to stderr."""
        self.log(click.style("Warning: %s" % msg, fg="yellow"), *args, file=sys.stderr)

    def elog(self, msg, *args):
        """Logs a error message to stderr."""
        self.log(click.style("Error: %s" % msg, fg="red"), *args, file=sys.stderr)

    def glog(self, msg, *args):
        """Logs a green message."""
        self.log(click.style("%s" % msg, fg="green"), *args)

    def vlog(self, msg, *args):
        """Logs a message to stderr only if verbose is enabled."""
        if self.verbose:
            self.log(msg, *args, **{"file": sys.stderr})

    def handle_error(self, exception):
        exc_type, exc_value, tb = sys.exc_info()
        if not self.debug:
            sys.stderr.write(f"\nError: {exc_value}, exception {exception}\n")
            sys.exit(1)
        else:
            reraise(exc_type, exc_value, tb.tb_next)

    def elapsed_time(self):
        return time.time() - self.t0

    def show_elapsed_time(self):
        duration = "{:.2f}".format(self.elapsed_time())
        self.vlog("Elapsed Time: " + (click.style(duration, fg="green")) + " seconds")

    def load_nxc(self, f):
        self.nxc = json.load(f)
        if "platform" in self.nxc and self.nxc["platform"] == "Grid5000":
            self.platform = Grid5000Platform(self)

    def load_role_quantities_file(self, filename):
        # (basename, extension)
        filename_tuple = os.path.splitext(filename)
        extension = filename_tuple[1]
        if extension in [".yaml", ".yml"]:
            self._load_role_quantities_file_yaml(filename)
        else:
            self._load_role_quantities_file_json(filename)

    def _load_role_quantities_file_json(self, filename):
        with open(filename, "r") as roles_f:
            self.set_roles_quantities(json.load(roles_f))

    def _load_role_quantities_file_yaml(self, filename):
        with open(filename, "r") as roles_f:
            self.set_roles_quantities(yaml.load(roles_f, Loader=get_nxc_loader()))

    def set_roles_quantities(self, roles_quantities):
        self.roles_quantities = roles_quantities


def make_pass_decorator(ensure=False):
    def decorator(f):
        @click.pass_context
        def new_func(*args, **kwargs):
            ctx = args[0]
            if ensure:
                obj = ctx.ensure_object(Context)
            else:
                obj = ctx.find_object(Context)
            try:
                return ctx.invoke(f, obj, *args[1:], **kwargs)
            except Exception as e:
                obj.handle_error(e)

        return update_wrapper(new_func, f)

    return decorator


class DeprecatedCmdDecorator(object):
    """This is a decorator which can be used to mark cmd as deprecated. It will
    result in a warning being emmitted when the command is invoked."""

    def __init__(self, message=""):
        if message:
            self.message = "%s." % message
        else:
            self.message = message

    def __call__(self, f):
        @click.pass_context
        def new_func(ctx, *args, **kwargs):
            msg = click.style(
                "warning: `%s` command is deprecated. %s"
                % (ctx.info_name, self.message),
                fg="yellow",
            )
            click.echo(msg)
            return ctx.invoke(f, *args, **kwargs)

        return update_wrapper(new_func, f)


class OnStartedDecorator(object):
    def __init__(self, callback):
        self.callback = callback
        self.exec_before = True

    def invoke_callback(self, ctx):
        if isinstance(self.callback, str):
            cmd = ctx.parent.command.get_command(ctx, self.callback)
            ctx.invoke(cmd)
        else:
            self.callback(ctx.obj)

    def __call__(self, f):
        @click.pass_context
        def new_func(ctx, *args, **kwargs):
            try:
                if self.exec_before:
                    self.invoke_callback(ctx)
                return ctx.invoke(f, *args, **kwargs)
            finally:
                if not self.exec_before:
                    self.invoke_callback(ctx)

        return update_wrapper(new_func, f)


class OnFinishedDecorator(OnStartedDecorator):
    def __init__(self, callback):
        super(on_finished, self).__init__(callback)
        self.exec_before = False


pass_context = make_pass_decorator(ensure=True)
on_started = OnStartedDecorator
on_finished = OnFinishedDecorator



import json
import os
import os.path as op
import glob
import socket
import sys
import shutil
import subprocess
import time
import base64
import click
import signal
import psutil
import itertools
import ipaddress
import urllib.request

from halo import Halo
from .tools.kataract import generate_scp_tasks, exec_kataract_tasks
from .default_role import DefaultRole


##
# Helper function to determine fstype
#
def get_fs_type(path):
    root_type = ""
    for part in psutil.disk_partitions(True):
        if part.mountpoint == "/":
            root_type = part.fstype
            continue
        if path.startswith(part.mountpoint):
            return part.fstype
    return root_type


##
# Retrieve from path from different store location if needed
#


def realpath_from_store(ctx, path):
    p = op.realpath(path)
    for store_path in ctx.alternative_stores:
        new_p = f"{store_path}{p[4:]}"
        if op.exists(new_p):
            return new_p
    if op.exists(p):
        return p
    ctx.elog(f"{path} does not exist in standard store or alternate")
    sys.exit(1)


##
# Generate/manipulate/copy deploy, compose files
#
def get_deployment_file(ctx, deployment_file):
    def exit_is_not_file(f):
        if not op.isfile(f):
            ctx.elog(f"{f} is not a file, deployment_file option must be provided")
            sys.exit(1)

    if not deployment_file:
        deploy_dir = op.join(ctx.envdir, "deploy")
        if not op.isdir(deploy_dir):
            ctx.elog("Failed to find deploy directory, is composition started ?")
            sys.exit(1)
        deployment_file = max(glob.glob(f"{deploy_dir}/*::*"), key=op.getctime)
        exit_is_not_file(deployment_file)
        return deployment_file
    else:
        base_deployment_file = deployment_file
        if op.exists(deployment_file):
            exit_is_not_file(deployment_file)
            return deployment_file
        else:
            deployment_file = op.join(op.join(ctx.envdir, "deploy"), deployment_file)
            if op.exists(deployment_file):
                exit_is_not_file(deployment_file)
                return deployment_file
            else:
                ctx.elog(f"{base_deployment_file} not found")
                sys.exit(1)


def read_deployment_info(ctx, deployment_file=None):
    ctx.deployment_filename = get_deployment_file(ctx, deployment_file)
    with open(ctx.deployment_filename, "r") as f:
        deployment_info = json.load(f)
    ctx.deployment_info = deployment_info
    return


def read_deployment_info_str(ctx, deployment_file=None):
    ctx.deployment_filename = get_deployment_file(ctx, deployment_file)
    with open(ctx.deployment_filename, "r") as f:
        deployment_info_str = f.read()
    return deployment_info_str


def read_test_script(ctx, compose_info_or_str):
    if isinstance(compose_info_or_str, str):
        filename = compose_info_or_str
    elif "test_script" in compose_info_or_str:
        filename = compose_info_or_str["test_script"]
    else:
        return None
    with open(realpath_from_store(ctx, filename), "r") as f:
        test_script = f.read()
        return test_script


def read_compose_info(ctx):

    if not op.isfile(ctx.compose_info_file):
        raise click.ClickException(f"{ctx.compose_info_filename} does not exist")
    with open(ctx.compose_info_file, "r") as f:
        compose_info = json.load(f)

    if "compositions_info" in compose_info:
        ctx.compositions_info = compose_info

        if len(compose_info["compositions_info"]) > 1:
            ctx.multiple_compositions = True
            ctx.vlog(
                f"Image with multiple compositions is detected, selected composition: {ctx.composition_name}"
            )

        if (
            ctx.composition_name
            and ctx.composition_name not in ctx.compositions_info["compositions_info"]
        ):
            ctx.elog(
                f'Composition named: "{ctx.composition_name}" is not in {ctx.compose_info_file}'
            )
            if ctx.composition_name == ctx.composition_basename_file:
                ctx.elog(
                    "When image contains multiple compositions and without default, the one to launch must by indicated via composition name option with name as prefix followed by file's label separated by ::  (e.g. -c foo::g5k-ramdisk). The default if present, is composition with its name equals to file's label."
                )
            sys.exit(1)
        if ctx.composition_name:
            compose_info = ctx.compositions_info["compositions_info"][
                ctx.composition_name
            ]

        if "all" not in compose_info:
            compose_info["all"] = ctx.compositions_info["all"]
            compose_info["flavour"] = ctx.compositions_info["flavour"]
            if "compositions_info_path" in ctx.compositions_info:
                compose_info["compositions_info_path"] = ctx.compositions_info[
                    "compositions_info_path"
                ]

        compose_flavour_name = ctx.compositions_info["flavour"]["name"]
        if ctx.flavour.name != compose_flavour_name:
            raise click.ClickException(
                f"Selected flavour ({ctx.flavour.name}) differs from compose info ({compose_flavour_name})"
            )

    ctx.compose_info = compose_info
    return


def read_hosts(hostsfile):
    return [host.rstrip() for host in open(hostsfile, "r")]


def translate_hosts2ip(ctx, hosts):
    for host in hosts:
        if host and (host not in ctx.host2ip_address):
            ip = socket.gethostbyname_ex(host)[2][0]
            ctx.host2ip_address[host] = ip
            ctx.ip_addresses.append(ip)
    return


def populate_deployment_vm_by_ip(nodes_info, roles_quantities):
    roles_quantities = health_check_roles_quantities(nodes_info, roles_quantities)
    i = 0
    deployment = {}
    ips = []
    for role, v in nodes_info.items():
        for hostname in roles_quantities[role]:
            ip = "10.0.2.{}".format(15 + i)
            ips.append(ip)
            # deployment[ip] = {"role": role, "vm_id": i}
            deployment[ip] = {
                "role": role,
                "init": v["init"],
                "vm_id": i,
                "host": hostname,
            }
            if "qemu_script" in v:
                deployment[ip]["qemu_script"] = v["qemu_script"]
            i = i + 1

    return deployment, ips


def health_check_roles_quantities(nodes_info, roles_quantities_in, ips=None):
    roles_quantities = {}
    remaining_role = None

    if len(roles_quantities_in) == 0:
        # If no info we take one node per role
        roles_quantities = {role: [role] for role in nodes_info.keys()}
    else:
        sum_nb_asked_machines = 0
        for role in roles_quantities_in:
            if type(roles_quantities_in[role]) == int:
                sum_nb_asked_machines += roles_quantities_in[role]
            elif type(roles_quantities_in[role]) == list:
                sum_nb_asked_machines += len(roles_quantities_in[role])
            else:
                pass
        if ips is not None:
            remaining_available_machines = len(ips) - sum_nb_asked_machines
        else:
            remaining_available_machines = -1

        # Step 1: if the user only gave the number of nodes of the roles
        for role in roles_quantities_in:
            if type(roles_quantities_in[role]) == int:
                nb_nodes = roles_quantities_in[role]
                if nb_nodes == 1:
                    roles_quantities[role] = [f"{role}"]
                else:
                    roles_quantities[role] = [
                        f"{role}{i}" for i in range(1, nb_nodes + 1)
                    ]
            elif type(roles_quantities_in[role]) == DefaultRole:
                nb_min_nodes = roles_quantities_in[role].nb_min_nodes
                if ips is None:
                    # in the case of VMs we are not limited
                    # so we take the min nb of nodes
                    remaining_available_machines = nb_min_nodes
                if remaining_available_machines < nb_min_nodes:
                    raise Exception(
                        f"Not enough nodes to satisfy default role {role} ({remaining_available_machines} available for {nb_min_nodes} asked)"
                    )
                if remaining_available_machines == 1:
                    roles_quantities[role] = [f"{role}"]
                else:
                    roles_quantities[role] = [
                        f"{role}{i}" for i in range(1, remaining_available_machines + 1)
                    ]
            else:
                if not ips:
                    raise Exception(
                        "Number of ip_address must be known with 'remaining' as role's number"
                    )
                if remaining_role:
                    raise Exception(
                        f"Role for remaining nodes is already set: {remaining_role}/{role}"
                    )
                else:
                    remaining_role = role

        # Step 2: add remainings and check that we do not have any conflict on the hostnames
        all_hostnames = list(itertools.chain.from_iterable(roles_quantities.values()))
        set_hostnames = set(all_hostnames)
        if len(all_hostnames) != len(set_hostnames):
            raise Exception("Conflict in the naming of the node")
    return roles_quantities


def populate_deployment_ips(nodes_info, ips, roles_quantities):
    roles_quantities = health_check_roles_quantities(nodes_info, roles_quantities, ips)
    i = 0
    deployment = {}
    for role, v in nodes_info.items():
        for hostname in roles_quantities[role]:
            ip = ips[i]
            deployment[ip] = {"role": role, "host": hostname, "init": v["init"]}
            i = i + 1
    return deployment


def populate_deployment_forward_ssh_port(nodes):
    i = 0
    deployment = {}
    for role in nodes:
        ip0 = 1 + i
        ip = f"127.0.0.{ip0}"
        deployment[ip] = {"host": role, "role": role, "ssh-port": 22022 + i}
        i = i + 1

    return deployment


def generate_deployment_info(ctx, ssh_pub_key_file=None):
    if not ctx.compose_info:
        read_compose_info(ctx)

    if not ssh_pub_key_file:
        ssh_pub_key_file = os.environ["HOME"] + "/.ssh/id_rsa.pub"
    with open(ssh_pub_key_file, "r") as f:
        sshkey_pub = f.read().rstrip()

    if ctx.forward_ssh_port and sshkey_pub:
        # QEMU_KERNEL_PARAMS environement variable is used to give ssh pub
        # key with nixos-test-ssh flavour (see boot.initrd.postMountCommands in nix/base.nix)
        qemu_kernel_params = ""
        if "QEMU_KERNEL_PARAMS" in os.environ:
            qemu_kernel_params = os.environ["QEMU_KERNEL_PARAMS"]
        ssh_key_pub_b64 = base64.b64encode(sshkey_pub.encode()).decode()
        os.environ[
            "QEMU_KERNEL_PARAMS"
        ] = f"{qemu_kernel_params} ssh_key.pub:{ssh_key_pub_b64}"

    # if ctx.multiple_compositions:  :: TO REMOVE ???
    #    nodes = ctx.compose_info["nodes"]

    if ctx.ip_addresses:
        deployment = populate_deployment_ips(
            ctx.compose_info["nodes"], ctx.ip_addresses, ctx.roles_quantities
        )
    elif ctx.forward_ssh_port:
        deployment = populate_deployment_forward_ssh_port(ctx.compose_info["nodes"])
    else:
        deployment, ctx.ip_addresses = populate_deployment_vm_by_ip(
            ctx.compose_info["nodes"], ctx.roles_quantities
        )
        deployment = {
            k: {
                "role": v["role"],
                "host": v["host"],
                "vm_id": v["vm_id"],
                "init": v["init"] if "host" in v else v["role"],
            }
            for k, v in deployment.items()
        }
    deployment = {
        "ssh_key.pub": sshkey_pub,
        "deployment": deployment,
    }

    if "all" in ctx.compose_info:
        deployment["all"] = ctx.compose_info["all"]

    if "compositions_info_path" in ctx.compose_info:
        deployment["compositions_info_path"] = ctx.compose_info[
            "compositions_info_path"
        ]

    if ctx.composition_name:
        deployment["composition"] = ctx.composition_name

    # Add user, used to determine nfs mount path on Grid'5000 by example
    # TODO: add option to override this (
    deployment["user"] = os.environ["USER"]

    # for k in ["all", "flavour"]:
    #    if k in compose_info:
    #        deployment[k] = compose_info[k]

    # If there is too much nodes httpd must used to tranfert deployment info,
    # due to kernel parameter size limit (deployment_info_b64 certainly will exceed it)
    if len(deployment["deployment"]) > 4:
        ctx.use_httpd = True

    json_deployment = json.dumps(deployment, indent=2)

    deploy_dir = op.join(ctx.envdir, "deploy")
    if not op.exists(deploy_dir):
        create = click.style("   create", fg="green")
        ctx.log("   " + create + "  " + deploy_dir)
        os.mkdir(deploy_dir)

    with open(
        op.join(deploy_dir, f"{ctx.composition_flavour_prefix}.json"), "w"
    ) as outfile:
        outfile.write(json_deployment)

    ctx.deployment_info = deployment
    return


def generate_kexec_scripts(ctx, flavour_kernel_params=""):
    if ctx.use_httpd:
        base_url = f"http://{ctx.httpd.ip}:{ctx.httpd.port}"
        deploy_info_src = f"{base_url}/deploy/{ctx.composition_flavour_prefix}.json"
    else:
        generate_deploy_info_b64(ctx)
        deploy_info_src = ctx.deployment_info_b64

    base_path = op.join(
        ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
    )
    kexec_scripts_path = op.join(base_path, "kexec_scripts")
    os.makedirs(kexec_scripts_path, mode=0o700, exist_ok=True)

    if "all" in ctx.deployment_info:
        kernel_path = realpath_from_store(ctx, ctx.deployment_info["all"]["kernel"])
        initrd_path = realpath_from_store(ctx, ctx.deployment_info["all"]["initrd"])

        kexec_args = "-l $KERNEL --initrd=$INITRD "
        kexec_args += f"--append='deploy={deploy_info_src} console=tty0 console=ttyS0,115200 {flavour_kernel_params} $DEBUG_INITRD'"
        script_path = op.join(kexec_scripts_path, "kexec.sh")
        with open(script_path, "w") as kexec_script:
            kexec_script.write("#!/usr/bin/env bash\n")
            kexec_script.write(": ${KERNEL:=" + kernel_path + "}\n")
            kexec_script.write(": ${INITRD:=" + initrd_path + "}\n")
            kexec_script.write(f"kexec {kexec_args}\n")
            kexec_script.write("kexec -e\n")
        os.chmod(script_path, 0o755)
    else:
        for ip, v in ctx.deployment_info["deployment"].items():
            role = v["role"]
            kernel_path = f"{base_path}/kernel_{role}"
            initrd_path = f"{base_path}/initrd_{role}"
            init_path = v["init"]
            kexec_args = f"-l {kernel_path} --initrd={initrd_path} "
            kexec_args += f"--append='init={init_path} deploy={deploy_info_src} console=tty0 console=ttyS0,115200 {flavour_kernel_params} $DEBUG_INITRD'"
            script_path = op.join(kexec_scripts_path, f"kexec_{role}.sh")
            with open(script_path, "w") as kexec_script:
                kexec_script.write("#!/usr/bin/env bash\n")
                kexec_script.write(f"kexec {kexec_args}\n")
                kexec_script.write("kexec -e\n")

            os.chmod(script_path, 0o755)


def generate_deploy_info_b64(ctx):

    deployment_info = {
        k: ctx.deployment_info[k]
        for k in [n for n in ctx.deployment_info.keys() if n != "deployment"]
    }

    deployment = {
        k: {"role": v["role"], "host": v["host"] if "host" in v else v["role"]}
        for k, v in ctx.deployment_info["deployment"].items()
    }

    # TODO: function to add multiple hostnames with same role
    # deployment = { k: {"role": v["role"], "host": "yopXXX"} for k,v in ctx.deployment_info["deployment"].items()}

    deployment_info["deployment"] = deployment

    ctx.vlog(f"deploy info \n{deployment_info}")
    deployment_info_str = json.dumps(deployment_info)

    ctx.deployment_info_b64 = base64.b64encode(deployment_info_str.encode()).decode()

    if len(ctx.deployment_info_b64) > (4096 - 256):
        ctx.log(
            "The base64 encoded deploy data is too large: use an http server to serve it"
        )
        sys.exit(1)
    return


# def copy_result_from_store(ctx):

#     if not ctx.compose_info:
#         read_compose_info(ctx)

#     if ctx.multiple_compositions:
#         composition_directory = "::"
#     else:
#         composition_directory = ctx.composition_name

#     artifact_path = op.join(
#         ctx.envdir, f"artifact/{composition_directory}/{ctx.flavour_name}"
#     )
#     if not op.exists(artifact_path):
#         create = click.style("   create", fg="green")
#         ctx.log("   " + create + "  " + artifact_path)
#         pathlib.Path(artifact_path).mkdir(parents=True, exist_ok=True)

#     if ctx.multiple_compositions:
#         compose_info = ctx.compositions_info
#     else:
#         compose_info = ctx.compose_info

#     new_compose_info = compose_info.copy()

#     if "all" in compose_info:
#         for target in ["kernel", "initrd", "qemu_script", "image"]:
#             if target in compose_info["all"]:
#                 new_target = op.join(artifact_path, target)
#                 shutil.copy(compose_info["all"][target], new_target)
#                 os.chmod(new_target, 0o644)
#                 new_compose_info["all"][target] = new_target
#     else:
#         for r, v in compose_info["nodes"].items():
#             for target in ["kernel", "initrd", "qemu_script", "image"]:
#                 if target in v:
#                     new_target = op.join(artifact_path, f"{target}_{r}")
#                     shutil.copy(v[target], new_target)
#                     os.chmod(new_target, 0o644)
#                     new_compose_info["nodes"][target] = new_target

#     if "test_script" in compose_info:
#         new_target = op.join(artifact_path, "test_script")
#         shutil.copy(compose_info["test_script"], new_target)
#         os.chmod(new_target, 0o644)
#         new_compose_info["test_script"] = new_target

#     # save new updated compose_info
#     json_new_compose_info = json.dumps(new_compose_info, indent=2)
#     with open(
#         op.join(ctx.envdir, f"build/{ctx.composition_flavour_prefix}::artifact"), "w",
#     ) as outfile:
#         outfile.write(json_new_compose_info)

#     ctx.compose_info = new_compose_info
#     ctx.log("Copy from store: " + click.style("completed", fg="green"))


##
#  Operation: connect, launch, wait_ssh_port,
#


def launch_ssh_kexec(ctx, ip=None, debug=False):
    if "all" in ctx.deployment_info:
        if ctx.push_path:
            kexec_script = f"{ctx.push_path}/kexec.sh"
            ki = f"KERNEL={ctx.push_path}kernel INITRD={ctx.push_path}initrd"
            user = "root@"
        else:
            base_path = op.join(
                ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
            )
            kexec_script = op.join(base_path, "kexec_scripts/kexec.sh")
            ki = ""
            user = ""
        # USELESS ? TOREMOVE ?
        # if ctx.sudo:
        #     sudo = f"SUDO={ctx.sudo}"
        # else:
        #     sudo = ""

        if "DEBUG_STAGE1" in os.environ or debug:
            # debug_stage1 = os.environ["DEBUG_STAGE1"]
            # TODO
            ctx.wlog(
                "Machine selection not yet supported within ssh_kexec. Debug_stage1 will apply to all."
            )
            ki = f" {ki} DEBUG_INITRD=boot.debug1mounts "

        def one_ssh_kexec(ip_addr):
            ssh_cmd = f'{ctx.ssh} {user}{ip_addr} "screen -dm bash -c \\" {ki} {kexec_script}\\""'
            ctx.vlog(ssh_cmd)
            subprocess.call(ssh_cmd, shell=True)

        if ip:
            one_ssh_kexec(ip)
        else:
            for ip in ctx.deployment_info["deployment"].keys():
                one_ssh_kexec(ip)
    else:
        raise Exception("Sorry, only all-in-one image version is supported up to now")


def wait_ssh_ports(ctx, ips=None, halo=True):
    ctx.log("Waiting ssh ports:")
    if not ips:
        ips = ctx.ip_addresses
    nb_ips = len(ips)

    nb_ssh_port = 0
    waiting_ssh_ports_cmd = (
        f"nmap -p22 -Pn {' '.join(ips)} -oG - | grep '22/open' | wc -l"
    )
    ctx.vlog(waiting_ssh_ports_cmd)
    if halo:
        spinner = Halo(text=f"Opened ssh ports 0/{nb_ips}", spinner="dots")
        spinner.start()
    while nb_ssh_port != nb_ips:
        output = subprocess.check_output(waiting_ssh_ports_cmd, shell=True)
        nb_ssh_port = int(output.rstrip().decode())
        if halo:
            spinner.text = "Opened ssh ports: {}/{} ({:.1f}s)".format(
                nb_ssh_port, nb_ips, ctx.elapsed_time()
            )
        time.sleep(0.25)
    if halo:
        spinner.succeed("Deployment taken {:.1f} sec".format(ctx.elapsed_time()))
    else:
        ctx.vlog("Deployment took {:.1f}s".format(ctx.elapsed_time()))


def push_on_machines(ctx):
    if "all" not in ctx.deployment_info:
        raise Exception("Sorry, only all-in-one image version is supported up to now")

    kernel = realpath_from_store(ctx, ctx.deployment_info["all"]["kernel"])
    initrd = realpath_from_store(ctx, ctx.deployment_info["all"]["initrd"])

    base_path = ctx.envdir

    # if ctx.multiple_compositions:
    #    subpath = "::"
    # else:
    # subpath = ctx.composition_name
    # base_path = op.join(
    #     ctx.envdir, f"artifact/{subpath}/{ctx.flavour_name}"
    # )
    subpath = ctx.composition_name
    base_path = op.join(ctx.envdir, f"artifact/{subpath}/{ctx.flavour_name}")

    kexec_script = op.join(base_path, "kexec_scripts/kexec.sh")

    ctx.vlog(
        f"push kernel, initrd, kexec_script on {ctx.ip_addresses} with scp executed concurrently"
    )
    for file_input in [kernel, initrd, kexec_script]:
        ctx.vlog(f"push: {file_input}")
        tasks_cmd = generate_scp_tasks(
            ctx.ip_addresses, file_input, ctx.push_path, scp="scp", user="root"
        )
        exec_kataract_tasks(tasks_cmd, elog=ctx.elog, vlog=ctx.vlog)

    # if shutil.which("kastafior"):
    #    raise NotImplementedError
    # elif shutil.which("kaput"):
    #     ctx.vlog("push kernel, initrd, kexec_script on hosts with kaput")
    #     joined_ip_addresses = ",".join(ctx.ip_addresses)
    #     for f in [kernel, initrd, kexec_script]:
    #         kaput_cmd = f"kaput -l root -n {joined_ip_addresses} {f} {ctx.push_path}"
    #         ctx.vlog(kaput_cmd)
    #         subprocess.call(kaput_cmd, shell=True)
    # else:
    #     for ip_address in ctx.ip_addresses:
    #         ctx.vlog(f"push kernel, initrd, kexec_script to {ip_address}")
    #         for f in [kernel, initrd, kexec_script]:
    #             subprocess.call(
    #                 f"scp {f} root@{ip_address}:{ctx.push_path}", shell=True
    #             )


def ssh_connect(ctx, user, node, execute=True):
    if not ctx.deployment_info:
        read_deployment_info(ctx)
    role = node
    host = node
    ssh_port = None
    for ip, v in ctx.deployment_info["deployment"].items():
        # if v["role"] == role:
        if v["host"] == role:
            host = ip
            if "ssh-port" in v:
                ssh_port = v["ssh-port"]
            break

    ssh_cmd = f"ssh -o StrictHostKeyChecking=no -o LogLevel=ERROR -l {user} "
    if ssh_port:
        ssh_cmd += f"-p {ssh_port} {host}"
    else:
        ssh_cmd += f"{host}"

    if execute:
        return_code = subprocess.run(ssh_cmd, shell=True).returncode

        if return_code:
            ctx.wlog(f"SSH exit code is not null: {return_code}")
        sys.exit(return_code)
    else:
        return ssh_cmd


NB_PANES_2_GEOMETRY = ["1", "1+1", "1+2", "2+2", "2+3", "3+3", "3+4", "4+4"]


def connect_tmux(ctx, user, nodes, pane_console, geometry, window_name="nxc"):
    if not nodes:
        deploy = ctx.deployment_info["deployment"]
        node = (list(deploy.keys()))[0]
        try:
            ipaddress.ip_address(node)
            nodes = [v["host"] for v in deploy.values()]
        except ValueError:
            nodes = list(deploy.keys())

    ssh_cmds = [ctx.flavour.ext_connect(user, node, execute=False) for node in nodes]

    console = 0
    if pane_console:
        console = 1

    if not geometry:
        geometry = ""
        nb_panes = len(nodes)
        if pane_console:
            nb_panes += 1

        if nb_panes < 9:
            geometry = NB_PANES_2_GEOMETRY[nb_panes - 1]

    # translate geometry
    if "+" in geometry and "*" in geometry:
        raise Exception("Mixing + and * in geometry is not supported")
    if "+" in geometry:
        splitw = [int(i) for i in geometry.split("+")]
        splitw.reverse()
    elif "*" in geometry:
        g = geometry.split("*")
        splitw = [int(g[1]) for i in range(int(g[0]))]
    else:
        splitw = [int(geometry)]

    nb_panes = sum(splitw)
    nb_splitv = len(splitw)

    ctx.vlog(f"geometry: {geometry}")
    ctx.vlog(f"splitw: {splitw}")
    ctx.vlog(f"nb_panes: {nb_panes}")

    # prepare commands
    base_cmds = ["bash" for i in range(nb_panes)]

    nb_ssh_cmds = len(ssh_cmds)
    if (nb_ssh_cmds + console) > nb_panes:
        ssh_cmds = ssh_cmds[: nb_panes - console]
    cmds = base_cmds[: nb_panes - nb_ssh_cmds] + ssh_cmds

    ctx.vlog(f"cmds: {cmds}")

    cmds.reverse()

    if "TMUX" not in os.environ:
        cmd = "tmux new -d"
        subprocess.call(cmd, shell=True)

    i = 0

    cmd = f"tmux new-window -n {window_name} -d"
    if not pane_console:
        cmd += f" {cmds[i]}"
        i += 1
    subprocess.call(cmd, shell=True)

    # cmd = f'tmux splitw -h -p 50 -t {window_name}.0 "{ssh_cmds[0]}"'
    # subprocess.call(cmd, shell=True)

    for v in range(nb_splitv):
        ratio_h = round(100.0 / (nb_splitv - v))
        if ratio_h != 100:
            cmd = f"tmux splitw -h -p {ratio_h} -t {window_name}.0 {cmds[i]}"
            pane0 = 1
            # print("vertical |", cmd)
            # print(round(100 * (1.0 / (nb_splitv - v))))
            subprocess.call(cmd, shell=True)
            i += 1
        else:
            pane0 = 0

        for h in range(splitw[v] - 1):
            ratio_v = round(100 * (1 - (1.0 / (splitw[v] - h))))
            cmd = f"tmux splitw -v -p {ratio_v} -t {window_name}.{h+pane0} {cmds[i]}"
            # print(cmd)
            subprocess.call(cmd, shell=True)
            i += 1

    cmd = f"tmux select-pane -t {window_name}.0"
    subprocess.call(cmd, shell=True)

    cmd = f"tmux select-window -t {window_name}"
    subprocess.call(cmd, shell=True)

    if "TMUX" not in os.environ:
        cmd = "tmux attach"
        subprocess.call(cmd, shell=True)


# TODO launch_vm(ctx, kexec_info, debug=False):
# TODO launch_vm_deploy(ctx, deployment, httpd_port=0, debug=False):
# NOT USED
def launch_vm(ctx, httpd_port=0, debug=False):

    if not op.exists("/tmp/kexec-qemu-vde1.ctl/ctl"):
        ctx.log("need sudo to create tap0 interface")
        subprocess.call("sudo true", shell=True)

    for ip, v in ctx.deployment_info["deployment"].items():
        qemu_script = v["qemu_script"]
        # cmd_qemu_script = f"DEPLOY='deploy=http://10.0.2.1:{httpd_port}/deployment.json' TAP=1"
        #
        cmd_qemu_script = "TAP=1"
        if httpd_port:
            cmd_qemu_script = " DEPLOY=1"
        if debug:
            cmd_qemu_script += " DEBUG_INITRD=boot.debug1mounts"

        cmd_qemu_script += " VM_ID={:02d} {} &".format(v["vm_id"], qemu_script)
        ctx.log("launch: {}".format(v["role"]))
        ctx.vlog(f"command: {cmd_qemu_script}")


# Helpers
def kill_proc_tree(
    pid, sig=signal.SIGTERM, include_parent=True, timeout=None, on_terminate=None
):
    """Kill a process tree (including grandchildren) with signal
    "sig" and return a (gone, still_alive) tuple.
    "on_terminate", if specified, is a callback function which is
    called as soon as a child terminates.
    From: https://psutil.readthedocs.io/en/latest/#kill-process-tree
    """
    assert pid != os.getpid(), "won't kill myself"
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    if include_parent:
        children.append(parent)
    for p in children:
        try:
            p.send_signal(sig)
        except psutil.NoSuchProcess:
            pass
    gone, alive = psutil.wait_procs(children, timeout=timeout, callback=on_terminate)
    return (gone, alive)


# test nix available
def get_nix_command(ctx):
    nix_cmd = ["nix"]

    if not shutil.which("nix"):
        local_bin_nix = "{os.environ['HOME']}/.local/bin/nix"
        if not op.exists(local_bin_nix):
            ctx.elog(
                "Nix not found, it can by installed in $HOME/.local/bin with command: nxc helper install-nix"
            )
            sys.exit(1)
        else:
            nix_cmd = [local_bin_nix]

    nix_cmd += ["--extra-experimental-features", "nix-command flakes"]
    return nix_cmd


# get Nix-static
def install_nix_static(
    version="2.10.3", archi="x86_64", local_bin_path=f"{os.environ['HOME']}/.local/bin"
):

    if not op.exists(local_bin_path):
        os.makedirs(local_bin_path)
    nix_path = op.join(local_bin_path, "nix")

    urllib.request.urlretrieve(
        f"https://gitlab.inria.fr/nixos-compose/nix-static/-/raw/main/bin/nix-{version}-{archi}-unknown-linux-musl",
        nix_path,
    )

    os.chmod(nix_path, 0o755)



import time
from typing import List
from .driver.machine import Machine


class Flavour(object):
    name: str
    external_connect: bool = False
    machines: List[Machine] = []

    def __init__(self, ctx):
        self.ctx = ctx

    def check(self, state="running"):
        self.ctx.wlog(f"Check not implement for flavour: {self.name}")
        return -1

    def wait_on_check(self, state="running", mode="all", period=0.5, round=5):
        for _ in range(round):
            if mode == "all" and self.check(state) == len(self.machines):
                return True
            elif mode == "any" and self.check(state) > 0:
                return
            time.sleep(period)
        return False



import os
import os.path as op
import tomlkit


def apply_setup(
    ctx,
    selected_setup,
    nix_flags,
    composition_file,
    composition_flavour,
    flavour,
    setup_param,
    filename="setup.toml",
):

    update_setup_file = False

    setup_file = op.join(ctx.envdir, filename)
    if not op.exists(setup_file):
        return (nix_flags, composition_file)

    setup_toml = tomlkit.loads(open(op.join(ctx.envdir, setup_file)).read())

    if (
        not selected_setup
        and "project" in setup_toml
        and "selected" in setup_toml["project"]
    ):
        ctx.wlog(
            "Detecting selected setup variant without asking it, removing it from setup file "
        )
        del setup_toml["project"]["selected"]
        update_setup_file = True

    # Handle setup parameters
    # First remove if any
    if "override-params" in setup_toml:
        del setup_toml["override-params"]
    # Second set override parmetes is needed
    if setup_param:
        override_params = tomlkit.table()
        setup_toml["override-params"] = override_params
        for param in setup_param:
            n, v = param.split("=")
            try:
                v = int(v)
            except ValueError:
                pass
            override_params.add(n, v)
        update_setup_file = True

    if update_setup_file:
        with open(setup_file, "w") as f:
            f.write(tomlkit.dumps(setup_toml))
            f.flush()
            os.fsync(f.fileno())

    if "options" in setup_toml:
        if not nix_flags and "nix-flags" in setup_toml["options"]:
            nix_flags = setup_toml["options"]["nix-flags"]
        if not composition_file and "composition-file" in setup_toml["options"]:
            composition_file = setup_toml["options"]["composition-file"]
        if not composition_flavour and "composition-flavour" in setup_toml["options"]:
            composition_flavour = setup_toml["options"]["composition-flavour"]
        if not flavour and "flavour" in setup_toml["options"]:
            flavour = setup_toml["options"]["flavour"]

    if selected_setup:
        if selected_setup not in setup_toml:
            ctx.elog("Missing asked setup variant: ${selected_setup}")
            sys.exit(1)
        else:
            if "options" in setup_toml[selected_setup]:
                if "nix-flags" in setup_toml[selected_setup]["options"]:
                    nix_flags = setup_toml[selected_setup]["options"]["nix-flags"]

                if "composition-file" in setup_toml[selected_setup]["options"]:
                    composition_file = setup_toml[selected_setup]["options"][
                        "composition-file"
                    ]

            if "project" not in setup_toml:
                project = tomlkit.table()
                setup_toml.add("selected", selected_setup)
            else:
                setup_toml["project"]["selected"] = selected_setup

            with open(setup_file, "w") as f:
                f.write(tomlkit.dumps(setup_toml))
                f.flush()
                os.fsync(f.fileno())

    ctx.setup = setup_toml

    return (nix_flags, composition_file, composition_flavour, flavour)



from pathlib import Path
import io
import os
import pty
import subprocess

from .logger import rootlog


class VLan:
    """This class handles a VLAN that the run-vm scripts identify via its
    number handles. The network's lifetime equals the object's lifetime.
    """

    nr: int
    socket_dir: Path

    process: subprocess.Popen
    slirpvde_process: subprocess.Popen
    pid: int
    fd: io.TextIOBase

    def __repr__(self) -> str:
        return f"<Vlan Nr. {self.nr}>"

    def __init__(self, nr: int, tmp_dir: Path, tap0: bool = False):
        self.nr = nr
        self.socket_dir = tmp_dir / f"vde{self.nr}.ctl"

        # TODO: don't side-effect environment here
        os.environ[f"QEMU_VDE_SOCKET_{self.nr}"] = str(self.socket_dir)

        rootlog.info("start vlan")
        pty_master, pty_slave = pty.openpty()

        vde_cmd = ["vde_switch"]

        if tap0:
            rootlog.info(
                f"starting VDE switch for network {self.nr}, with tap0 (sudo required)"
            )
            subprocess.call("sudo true", shell=True)
            vde_cmd = ["sudo"] + vde_cmd + ["-tap", "tap0"]

        vde_cmd = vde_cmd + ["-s", self.socket_dir, "-mod", "0770", "-group", "users"]

        self.process = subprocess.Popen(
            vde_cmd,
            stdin=pty_slave,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
        )
        self.pid = self.process.pid
        self.fd = os.fdopen(pty_master, "w")
        self.fd.write("version\n")

        # TODO: perl version checks if this can be read from
        # an if not, dies. we could hang here forever. Fix it.
        assert self.process.stdout is not None
        self.process.stdout.readline()
        if not (self.socket_dir / "ctl").exists():
            rootlog.error("cannot start vde_switch")

        # setup tap0
        if tap0:
            rootlog.info("setup tap0 interface")
            # TODO add error handling
            subprocess.call("sudo ip addr add 10.0.2.1/24 dev tap0", shell=True)
            subprocess.call("sudo ip link set dev tap0 up", shell=True)

            # launch slirp
            rootlog.info(f"slirpvde -d -s {self.socket_dir} -dhcp -q")
            # subprocess.call(f'slirpvde -d -s {vde_socket} -dhcp -q', shell=True)
            slirpvde_process = subprocess.Popen(
                f"slirpvde -d -s {self.socket_dir} -dhcp -q",
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                shell=True,
            )
            self.slirpvde_process = slirpvde_process
            rootlog.info(f"running slirpvde (pid {slirpvde_process.pid})")

        rootlog.info(f"running vlan (pid {self.pid})")

    def __del__(self) -> None:
        rootlog.info(f"kill vlan (pid {self.pid})")
        self.fd.close()
        self.process.terminate()
        if self.slirpvde_process:
            self.slirpvde_process.terminate()
